import time
from threading import Event, Thread

import numpy as np

from kamzik3 import DeviceError
from kamzik3.constants import *
from kamzik3.devices.dummy.deviceDummy import DeviceDummy
from kamzik3.snippets.snippetsDecorators import expose_method
from kamzik3.snippets.snippetsUnits import device_units

ATTR_ACQUIRED_FRAMES = "Acquired frames"
ATTR_FRAME_NAME = u"Frame name"
ATTR_FRAME_FILE = u"Frame file"


class DummyDetector(DeviceDummy):

    def __init__(self, device_id=None, config=None):
        self.acquisition_running = Event()
        DeviceDummy.__init__(self, device_id, config)

    def _init_attributes(self):
        DeviceDummy._init_attributes(self)
        self.create_attribute(ATTR_EXPOSURE_TIME, default_value=0, readonly=False, default_type=np.float, unit="ms",
                              min_value=1)
        self.create_attribute(ATTR_FRAME_IMAGE, default_value=0, readonly=True, default_type=np.uint64)
        self.create_attribute(ATTR_FRAME_NAME, default_value="Frame", readonly=False)
        self.create_attribute(ATTR_FRAME_FILE, default_value=0, readonly=True)
        self.create_attribute(ATTR_SUMMED_FRAMES, default_value=0, readonly=False, default_type=np.uint64, min_value=1)
        self.create_attribute(ATTR_ACQUIRED_FRAMES, default_value=0, readonly=True, default_type=np.uint64)

    @expose_method()
    def stop(self):
        self.acquisition_running.clear()

    @expose_method({"Name": ATTR_FRAME_NAME, "Frames": ATTR_SUMMED_FRAMES, "Exposure": ATTR_EXPOSURE_TIME})
    def start_acquisition(self, Name, Frames, Exposure):
        if self.acquisition_running.isSet():
            raise DeviceError("Another acquisition is already running")
        self.acquisition_running.set()
        Frames = device_units(self, ATTR_SUMMED_FRAMES, Frames)
        Exposure = device_units(self, ATTR_EXPOSURE_TIME, Exposure)

        self.set_status(STATUS_BUSY)
        self.set_value(ATTR_FRAME_NAME, Name)
        self.set_value(ATTR_SUMMED_FRAMES, Frames)
        self.set_value(ATTR_EXPOSURE_TIME, Exposure)
        self.set_value(ATTR_ACQUIRED_FRAMES, 0)

        def _acquire():
            exposure_time_increment = 100
            current_exposure_time = 0

            while self.acquisition_running.isSet():
                time.sleep(exposure_time_increment * 1e-3)
                current_exposure_time += exposure_time_increment
                acquired_frames = int(current_exposure_time / Exposure.m)

                if acquired_frames >= Frames:
                    self.set_value(ATTR_ACQUIRED_FRAMES, Frames)
                    break
                else:
                    self.set_value(ATTR_ACQUIRED_FRAMES, acquired_frames)

            if self.connected:
                self.acquisition_running.clear()
                self.set_status(STATUS_IDLE)

        Thread(target=_acquire).start()



import logging
import time
from threading import Event

from kamzik3 import units, MacroException
from kamzik3.constants import *
from kamzik3.snippets.snippetsYaml import YamlSerializable


class Common(YamlSerializable):
    state = IDLE
    repeated = 0
    error_message = None
    previous_running_time = 0
    running_time = 0
    started_at = 0
    done_at = 0
    stopped_at = 0
    reset_at = 0
    error_at = 0
    steps_count = 0
    points_count = 1
    retry_timeout = 100  # 100 ms
    common_id = None
    logger = None

    def __init__(self, common_id, repeat_count=0, wait_after=units.Quantity(0, "s"), retry_count=0, trigger_log=True):
        assert repeat_count >= 0, u"Repeat parameter should be >= 0"
        assert wait_after >= 0, u"Wait_after parameter should be >= 0"
        self.set_common_id(common_id)
        self.max_retry_count = retry_count
        self.retry_count = retry_count
        self.wait_after = wait_after.to("s")
        self.trigger_log = trigger_log

        self.timer = None
        if self.wait_after > 0:
            self.timer = Event()
        self.repeat_count = int(repeat_count)

    def set_common_id(self, common_id):
        self.common_id = common_id
        if not hasattr(self, "logger") or self.logger is None:
            self.logger = logging.getLogger("Macro.common.{}".format(self.common_id))

    def set_state(self, new_state):
        self.state = new_state

    def get_state(self):
        return self.state

    def reset(self):
        self.logger.debug("Reset")
        self.reset_at = time.time()
        self.error_message = None
        self.repeated = 0
        self.retry_count = self.max_retry_count
        self.set_state(IDLE)

    def start(self):
        self.logger.debug("Started")
        self.started_at = time.time()
        self.running_time = 0
        self.set_state(BUSY)

    def repeat(self):
        self.logger.debug("Repeated")
        self.started_at = time.time()
        self.running_time = 0
        self.repeated += 1
        self.set_state(BUSY)

    def retry(self, retry_reason):
        if self.retry_count == 0:
            raise retry_reason
        else:
            self.retry_count -= 1
            self.logger.error(
                "Retrying step due to error. Remaining tries {}. Error message: {}".format(self.retry_count,
                                                                                           retry_reason))
            self.reset_at = time.time()
            self.error_message = None
            time.sleep(self.retry_timeout / 1e3)

    def stop(self):
        self.stopped_at = time.time()
        self.running_time = self.stopped_at - self.started_at
        if self.timer is not None:
            self.stop_timer()
        self.logger.debug("Stopped")
        self.set_state(STOPPED)

    def done(self):
        self.done_at = time.time()
        self.previous_running_time = self.running_time = self.done_at - self.started_at
        self.logger.debug("Done")
        self.set_state(DONE)

    def error(self, error_message):
        self.error_at = time.time()
        self.running_time = self.error_at - self.started_at
        self.error_message = error_message
        self.logger.error(error_message)
        self.set_state(ERROR)

    def stop_timer(self):
        if self.timer is not None:
            self.logger.debug("Timer for {}s stopped".format(self.wait_after))
            self.timer.set()
            return True
        return False

    def start_timer(self):
        if self.timer is not None:
            self.logger.debug("Waiting for {}s".format(self.wait_after))
            self.timer.clear()
            self.timer.wait(self.wait_after.m)
            return True
        return False

    def get_total_steps_count(self):
        return self.steps_count * (self.repeat_count + 1)

    def remove(self):
        self.logger = None

    def get_reset_step(self):
        return None

    def get_output(self):
        return None


class StepGenerator(Common):
    as_step = False
    step_index = 0
    first_step_at = 0

    def return_back_callback(self):
        raise NotImplementedError()

    def step_generator(self):
        raise NotImplementedError()

    def reset(self):
        self.steps = self.step_generator()
        Common.reset(self)

    def repeat(self):
        Common.repeat(self)
        self.steps = self.step_generator()
        return self.start()

    def stop(self):
        self.started_at = self.first_step_at
        if self.current_step is not None:
            self.current_step.stop()
        Common.stop(self)

    def done(self):
        self.started_at = self.first_step_at
        Common.done(self)

    def error(self, error_message):
        self.started_at = self.first_step_at
        Common.error(self, error_message)

    def before_first_step_executed(self):
        pass

    def before_step_executed(self):
        pass

    def start(self):
        Common.start(self)
        try:
            if self.current_step is not None:
                self.current_step.remove()
            else:
                self.before_first_step_executed()
            self.current_step = next(self.steps)
            self.before_step_executed()
            return self.current_step
        except StopIteration:
            self.done()
            return None
        except MacroException as e:
            self.error(e)
            raise e



import kamzik3
from kamzik3.devices.attribute import Attribute
from kamzik3.devices.device import Device


class MeasurementGroup(list):

    def __init__(self, group_id, measurement_group=None):
        self.group_id = group_id
        if measurement_group is not None:
            list.__init__(self)
            for device_id, attribute in measurement_group:
                self.append(device_id, attribute)
        else:
            list.__init__(self)
        kamzik3.session.add_measurement_group(self)

    def append(self, device_id, attribute):
        if isinstance(device_id, Device):
            device_id = device_id.device_id
        item = (device_id, Attribute.list_attribute(attribute))
        if item not in self:
            list.append(self, item)



import time

import numpy as np
import serial

from kamzik3.constants import *
from kamzik3.devices.devicePort import DevicePort
from kamzik3.devices.deviceSocket import DeviceSocket


class DevicePeakTech(DeviceSocket):
    terminator = b"OK\r"
    push_commands_max = 1

    def _init_attributes(self):
        super(DevicePeakTech, self)._init_attributes()
        self.create_attribute(ATTR_CONSTANT_CURRENT, readonly=True, default_type=np.bool)
        self.create_attribute(ATTR_CONSTANT_VOLTAGE, readonly=True, default_type=np.bool)
        self.create_attribute(ATTR_VOLTAGE, default_type=np.float16, min_value=0, decimals=3, unit="V",
                              set_function=self.set_voltage_output)
        self.create_attribute(ATTR_MAX_VOLTAGE, default_type=np.float16, min_value=0, decimals=1, unit="V",
                              max_value=99.9, set_function=self.set_max_voltage_output,
                              set_value_when_set_function=False)
        self.create_attribute(ATTR_CURRENT, default_type=np.float16, min_value=0, decimals=3, unit="A",
                              set_function=self.set_current_output)
        self.create_attribute(ATTR_MAX_CURRENT, default_type=np.float16, min_value=0, max_value=99.9, decimals=1,
                              unit="A", set_function=self.set_max_current_output, set_value_when_set_function=False)
        self.create_attribute(ATTR_OUTPUT_SWITCH, default_type=np.bool, set_function=self.switch_output,
                              set_value_when_set_function=False)

    def handle_configuration(self):
        start_at = time.time()

        def _finish_configuration(*_, **__):
            self._config_commands()
            self._config_attributes()
            self.start_polling()
            self.set_status(STATUS_CONFIGURED)
            self.logger.info(u"Device configuration took {} sec.".format(time.time() - start_at))

        self.command(u"GOVP\r")
        self.command(u"GOCP\r", callback=_finish_configuration)

    def handle_readout(self, readout_buffer):
        command, output, callback, token = super(DevicePeakTech, self).handle_readout(readout_buffer)

        if command == u"GETD\r":
            voltage, current, cv_cc_flag = int(output[:4]) / 100., int(output[4:8]) / 100, int(output[8:])
            self.set_raw_value(ATTR_CONSTANT_VOLTAGE, True if cv_cc_flag == 0 else False)
            self.set_raw_value(ATTR_CONSTANT_CURRENT, True if cv_cc_flag == 1 else False)
            self.set_raw_value(ATTR_VOLTAGE, voltage)
            self.set_raw_value(ATTR_CURRENT, current)
            if voltage < 0.75:
                self.set_raw_value(ATTR_OUTPUT_SWITCH, False)
            else:
                self.set_raw_value(ATTR_OUTPUT_SWITCH, True)
        elif command == u"GOVP\r":
            max_voltage = int(output[:3]) / 10
            self.set_attribute((ATTR_VOLTAGE, MAX), max_voltage)
            self.set_raw_value(ATTR_MAX_VOLTAGE, max_voltage)
        elif command == u"GOCP\r":
            max_current = int(output[:3]) / 10
            self.set_attribute((ATTR_CURRENT, MAX), max_current)
            self.set_raw_value(ATTR_MAX_CURRENT, max_current)

        if callback is not None:
            self.handle_readout_callback(callback, command, readout_buffer)
        if token:
            self.device_server.push_message(self.device_id, [command, readout_buffer], token)

    def start_polling(self):
        super().start_polling()
        self.poll_command(u"GETD\r", 200)

    def set_max_voltage_output(self, value, callback=None):
        self.logger.info(u"Set max voltage output to {} V".format(value))
        self.command(u"SOVP{0:03d}\r".format(int(float(value) * 10.)), callback, with_token=True)
        return self.command("GOVP\r")

    def set_max_current_output(self, value, callback=None):
        self.logger.info(u"Set max current output to {} A".format(value))
        self.command(u"SOCP{0:03d}\r".format(int(float(value) * 10.)), callback, with_token=True)
        return self.command("GOCP\r")

    def set_voltage_output(self, value, callback=None):
        self.logger.info(u"Set voltage output to {} V".format(value))
        return self.command(u"VOLT{0:03d}\r".format(int(float(value) * 10.)), callback, with_token=True)

    def set_current_output(self, value, callback=None):
        self.logger.info(u"Set current output to {} A".format(value))
        return self.command(u"CURR{0:03d}\r".format(int(float(value) * 10.)), callback, with_token=True)

    def switch_output(self, value, callback=None):
        self.logger.info(u"Switch output to {}".format(value))
        return self.command(u"SOUT{}\r".format(int(not value)), callback, with_token=True)

    def stop(self):
        self.set_voltage_output(0)
        self.set_current_output(0)


class DevicePeakTechPort(DevicePort, DevicePeakTech):
    terminator = DevicePeakTech.terminator
    push_commands_max = DevicePeakTech.push_commands_max

    def __init__(self, port, baud_rate=9600, parity=serial.PARITY_NONE, stop_bits=serial.STOPBITS_ONE,
                 byte_size=serial.EIGHTBITS, device_id=None, config=None):
        super().__init__(port, baud_rate, parity, stop_bits, byte_size, device_id, config)

    def handshake(self):
        self.logger.info(u"Handshake initiated")
        self.push(b"GETD\r")
        time.sleep(0.1)  # Give some time for devices to reply
        if self.serial_port.in_waiting != 0:
            readout = self.read_all().split("\r")
            if readout[1] == "OK":
                return True
        return False



from PyQt5.QtCore import pyqtSlot

from kamzik3.constants import *
from kamzik3.devices.attribute import Attribute
from kamzik3.gui.attributeDeviceDisplayWidget import AttributeDeviceDisplayWidget
from kamzik3.gui.deviceWidget import DeviceWidget
from kamzik3.gui.general.stackableWidget import StackableWidget
from kamzik3.gui.templates.deviceDisplayAttributeTemplate import Ui_Form


class DeviceDisplayAttributeWidget(Ui_Form, DeviceWidget, StackableWidget):
    attribute = None
    debug_widget = None
    input_value = None

    def __init__(self, device, attribute, config=None, parent=None):
        self.attribute = Attribute.list_attribute(attribute)
        DeviceWidget.__init__(self, device, config=config, parent=parent)
        self.set_status_label(self.device.get_value(ATTR_STATUS))
        self.label_device_name.setText(self.device.device_id)

    @pyqtSlot()
    def set_status_label(self, status):
        self.label_status.setText(status)
        self.label_status.setStyleSheet("QLabel {{background:{}}}".format(self.get_status_color(status)))

    @pyqtSlot()
    def slot_handle_configuration(self):
        DeviceWidget.slot_handle_configuration(self)

        self.input_not_ready_placeholder.setParent(None)
        layout = self.widget_holder.layout()
        self.input_value = AttributeDeviceDisplayWidget(self.device_id, self.attribute, config=self.config)
        layout.insertWidget(2, self.input_value)

    @pyqtSlot("bool")
    def slot_set_enabled(self, value):
        self.self.input_value.setEnabled(value)

    @pyqtSlot("PyQt_PyObject")
    def slot_set_status(self, value):
        DeviceWidget.slot_set_status(self, value)

        if value in READY_DEVICE_STATUSES:
            if self.input_value is not None:
                self.input_value.setEnabled(True)
        else:
            if self.input_value is not None:
                self.input_value.setEnabled(False)

        self.set_status_label(value)

    def close(self):
        self.input_value.close()
        self.input_value = None
        self.attribute = None
        self.input_value = None
        super(DeviceDisplayAttributeWidget, self).close()



import math

from PyQt5.QtWidgets import QWidget, QHBoxLayout, QButtonGroup

import kamzik3
from kamzik3 import units
from kamzik3.constants import *
from kamzik3.devices.attribute import Attribute
from kamzik3.gui.attributeDeviceDisplayWidget import AttributeDeviceDisplayWidget
from kamzik3.gui.deviceDebugWidget import DeviceDebugWidget
from kamzik3.gui.general.lockButtonWidget import LockButton
from kamzik3.gui.general.macroWidget import MacroWidget
from kamzik3.gui.templates.scanAttributeTemplate import Ui_Form
from kamzik3.macro.scan import Scan
from kamzik3.macro.step import StepDeviceAttributeNumerical
from kamzik3.snippets.snippetsWidgets import show_error_message, show_centralized_top


class ScanAttributeWidget(Ui_Form, MacroWidget):
    device = None
    attribute = None
    debug_widget = None

    def __init__(self, title=None, config=None, parent=None):
        MacroWidget.__init__(self, parent)
        self.config = config
        self.title = title
        if self.config is None:
            self.config = {}
        self.setupUi(self)

    def setupUi(self, Form):
        super().setupUi(Form)
        self.button_settings.setEnabled(False)

        if self.config.get("input", "relative") == "relative":
            self.radio_relative_scan.setChecked(True)
        else:
            self.radio_absolute_scan.setChecked(True)

        self.end_point_lock_button = LockButton()
        self.end_point_set_widget = QWidget()
        lock_layout = QHBoxLayout(self.end_point_set_widget)
        lock_layout.setContentsMargins(0, 0, 0, 0)
        lock_layout.addWidget(self.end_point_lock_button)

        self.total_steps_lock_button = LockButton()
        self.total_steps_set_widget = QWidget()
        lock_layout = QHBoxLayout(self.total_steps_set_widget)
        lock_layout.setContentsMargins(0, 0, 0, 0)
        lock_layout.addWidget(self.total_steps_lock_button)

        self.step_size_lock_button = LockButton()
        self.step_size_set_widget = QWidget()
        lock_layout = QHBoxLayout(self.step_size_set_widget)
        lock_layout.setContentsMargins(0, 0, 0, 0)
        lock_layout.addWidget(self.step_size_lock_button)

        self.field_size_lock_button = LockButton()
        self.field_size_set_widget = QWidget()
        lock_layout = QHBoxLayout(self.field_size_set_widget)
        lock_layout.setContentsMargins(0, 0, 0, 0)
        lock_layout.addWidget(self.field_size_lock_button)

        self.button_group = QButtonGroup()
        self.button_group.setExclusive(True)
        self.total_steps_lock_button.setDefault(True)
        self.total_steps_lock_button.setChecked(True)
        self.button_group.addButton(self.end_point_lock_button)
        self.button_group.addButton(self.total_steps_lock_button)
        self.button_group.addButton(self.step_size_lock_button)
        self.button_group.addButton(self.field_size_lock_button)

        self.spacer = self.gridLayout.itemAtPosition(5, 0)
        self.set_title(self.title)
        self._init_device_list()
        self._init_scanner_list()

        if self.config.get("attribute") is not None:
            self.label_attribute.hide()
            self.combobox_attribute.hide()

        self._init_template_settings()

    def _init_template_settings(self):
        if not self.config.get(CFG_TEMPLATE, False):
            self.label_settings.hide()
            self.widget_settings.hide()
            self.label_scanner.hide()
            self.combobox_scanner.hide()

    def _init_device_list(self):
        self.combobox_device.blockSignals(True)
        if self.config.get("attribute") is not None:
            attribute = self.config.get("attribute").split(u" # ")
            devices_list = list(kamzik3.session.get_devices(attribute_filter=attribute).keys())
        else:
            devices_list = list(kamzik3.session.get_devices().keys())

        self.combobox_device.addItems(["None"] + devices_list)
        self.combobox_device.blockSignals(False)

        if self.config.get("device_id") is not None:
            self.combobox_device.setCurrentText(self.config.get("device_id"))

    def _init_scanner_list(self):
        self.combobox_scanner.blockSignals(True)
        devices_list = list(kamzik3.session.get_devices(method_filter="get_scanner_macro").keys())

        self.combobox_scanner.addItems(["None"] + devices_list)
        self.combobox_scanner.blockSignals(False)

        if self.config.get("device_id") is not None:
            self.combobox_scanner.setCurrentText(self.config.get("device_id"))

    def _init_attribute_list(self, device):
        current_text = self.combobox_attribute.currentText()
        self.combobox_attribute.clear()
        self.combobox_attribute.blockSignals(True)
        if device is None:
            self.combobox_attribute.addItem("None")
        else:
            attribute_list = []
            self.get_scannable_attribute_list(device.attributes, attribute_list)
            if len(attribute_list) == 0:
                attribute_list.append("None")
            self.combobox_attribute.addItems(attribute_list)
        index = self.combobox_attribute.findText(current_text)
        if index == -1:
            self.combobox_attribute.setCurrentIndex(0)
        else:
            self.combobox_attribute.setCurrentIndex(index)
        self.combobox_attribute.blockSignals(False)
        self.attribute_selected(self.combobox_attribute.currentText())

    def get_scannable_attribute_list(self, attributes, output=[]):
        for k, v in attributes.items():
            if Attribute.is_attribute(v):
                if not v[READONLY] and v.numerical:
                    output.append(k)
            else:
                leaf = []
                self.get_scannable_attribute_list(v, leaf)
                output += map(lambda nk: "{} # {}".format(k, nk), leaf)

    def device_selected(self, device_id):
        if device_id == "None":
            self.device = None
            self._init_attribute_list(None)
            self.button_settings.setEnabled(False)
        else:
            self.button_settings.setEnabled(True)
            self.device = kamzik3.session.get_device(device_id)
            if self.config.get("attribute") is not None:
                self.attribute_selected(self.config.get("attribute"))
            else:
                self._init_attribute_list(self.device)

    def attribute_selected(self, attribute):
        if self.config.get(CFG_TEMPLATE, False):
            return

        for widget in self.macro_widgets:
            widget.close()
            widget.setParent(None)
        self.macro_widgets = []
        self.macro_inputs = {}

        # Remove all locked widgets
        self.end_point_set_widget.setParent(None)
        self.field_size_set_widget.setParent(None)
        self.step_size_set_widget.setParent(None)
        self.total_steps_set_widget.setParent(None)
        if attribute in ("", "None", None):
            # Don't show any widget if attribute is empty
            self.attribute = None
        else:
            attribute = attribute.split(u" # ")
            user_unit = AttributeDeviceDisplayWidget.user_attribute_units.get((self.device.device_id, tuple(attribute)),
                                                                              None)
            self.attribute = self.device.get_attribute(attribute)
            row = 3
            self.macro_widgets.append(self._show_total_steps_widget(row))
            row += 1
            self.macro_widgets.append(self._show_step_size_widget(row, user_unit))
            row += 1
            if self.radio_absolute_scan.isChecked():
                self.macro_widgets.append(self._show_start_point_widget(row, user_unit))
                row += 1
                self.macro_widgets.append(self._show_end_point_widget(row, user_unit))
            if self.radio_relative_scan.isChecked():
                row += 1
                self.macro_widgets.append(self._show_field_size_widget(row, user_unit))
            row += 1
            self.macro_widgets.append(self._show_negative_tolerance_widget(row, user_unit))
            row += 1
            self.macro_widgets.append(self._show_positive_tolerance_widget(row, user_unit))
            row += 1
            # self.macro_widgets.append(self._show_random_deviation_widget(row))
            # row += 1
            self.macro_widgets.append(self._show_wait_after_widget(row))
            row += 1
            self.macro_widgets.append(self._show_repeat_widget(row))
            row += 1
            self.macro_widgets.append(self._show_retry_widget(row))
            row += 1

            if self.config.get("scanner", "None") != "None":
                scanner = kamzik3.session.get_device(self.config["scanner"])
                try:
                    scanner_attributes = scanner.get_scanner_attributes()
                    for attribute in scanner_attributes:
                        self.macro_widgets.append(
                            self._show_method_attribute_widget(attribute, scanner.get_attribute(attribute).attribute_copy(),
                                                               row))
                        row += 1
                except AttributeError:
                    pass

            self.layout().removeItem(self.spacer)
            self.layout().addItem(self.spacer, row, 0)
            if self.radio_relative_scan.isChecked():
                self.signal_total_steps_changed(None)
            else:
                self.signal_absolute_total_steps_changed(None)

    """
    Signals and checks for absolute scan
    """

    def signal_start_end_changed(self, _):
        values = self.get_input_values()
        self.macro_inputs["step_size"].input_widget.blockSignals(True)
        self.macro_inputs["total_steps"].input_widget.blockSignals(True)
        field_size = values["end_point"] - values["start_point"]

        if self.step_size_lock_button.isChecked():
            try:
                self.macro_inputs["total_steps"].set_value(
                    units.Quantity(math.ceil(field_size / values["step_size"]))
                )
            except ZeroDivisionError:
                self.macro_inputs["total_steps"].set_value(units.Quantity(0))
        else:
            try:
                self.macro_inputs["step_size"].set_value(field_size / values["total_steps"])
            except ZeroDivisionError:
                self.macro_inputs["step_size"].set_value(units.Quantity(0, values["step_size"].u))

        self.macro_inputs["step_size"].input_widget.blockSignals(False)
        self.macro_inputs["total_steps"].input_widget.blockSignals(False)

    def signal_start_end_check_limits(self):
        values = self.get_input_values()
        if values["total_steps"] == 0 and values["step_size"] != 0:
            self.macro_inputs["end_point"].set_value(values["start_point"])

    def signal_absolute_step_size_changed(self, _):
        values = self.get_input_values()
        self.macro_inputs["end_point"].input_widget.blockSignals(True)
        self.macro_inputs["total_steps"].input_widget.blockSignals(True)
        field_size = values["end_point"] - values["start_point"]

        if self.total_steps_lock_button.isChecked():
            self.macro_inputs["end_point"].set_value(
                values["start_point"] + values["step_size"] * values["total_steps"])
        else:
            try:
                self.macro_inputs["total_steps"].set_value(
                    units.Quantity(math.ceil(field_size / values["step_size"]))
                )
            except ZeroDivisionError:
                self.macro_inputs["total_steps"].set_value(units.Quantity(0))

        self.macro_inputs["total_steps"].input_widget.blockSignals(False)
        self.macro_inputs["end_point"].input_widget.blockSignals(False)

    def signal_absolute_step_size_check_limits(self):
        values = self.get_input_values()
        field_size = values["end_point"] - values["start_point"]
        computed_end_point = values["start_point"] + (values["step_size"] * values["total_steps"])
        if abs(values["step_size"]) > abs(field_size):
            self.macro_inputs["step_size"].set_value(field_size)
        elif computed_end_point > (self.macro_inputs["end_point"].get_maximum() + values["step_size"]):
            try:
                self.macro_inputs["step_size"].set_value(
                    (self.macro_inputs["end_point"].get_maximum() - values["start_point"]) / values[
                        "total_steps"]
                )
            except ZeroDivisionError:
                self.macro_inputs["step_size"].set_value(units.Quantity(0, values["step_size"].u))
        elif computed_end_point == values["start_point"]:
            self.macro_inputs["step_size"].set_value(units.Quantity(0, values["step_size"].u))
        elif computed_end_point < (self.macro_inputs["end_point"].get_minimum() + values["step_size"]):
            self.macro_inputs["step_size"].set_value(field_size)

    def signal_absolute_total_steps_changed(self, _):
        values = self.get_input_values()
        self.macro_inputs["end_point"].input_widget.blockSignals(True)
        self.macro_inputs["step_size"].input_widget.blockSignals(True)
        field_size = values["end_point"] - values["start_point"]
        if self.step_size_lock_button.isChecked():
            self.macro_inputs["end_point"].set_value(
                values["start_point"] + values["step_size"] * values["total_steps"])
        else:
            try:
                self.macro_inputs["step_size"].set_value(field_size / values["total_steps"])
            except ZeroDivisionError:
                self.macro_inputs["step_size"].set_value(units.Quantity(0, values["step_size"].u))
        self.macro_inputs["step_size"].input_widget.blockSignals(False)
        self.macro_inputs["end_point"].input_widget.blockSignals(False)

    def signal_absolute_total_steps_check_limits(self):
        values = self.get_input_values()
        field_size = values["end_point"] - values["start_point"]
        if (values["total_steps"] * values["step_size"]) > self.macro_inputs["end_point"].get_maximum():
            self.macro_inputs["total_steps"].set_value(
                units.Quantity(math.ceil(field_size / values["step_size"]))
            )

    def signal_field_size_changed(self, __):
        values = self.get_input_values()
        self.macro_inputs["step_size"].input_widget.blockSignals(True)
        self.macro_inputs["total_steps"].input_widget.blockSignals(True)
        if self.total_steps_lock_button.isChecked():
            try:
                self.macro_inputs["step_size"].set_value(
                    values["field_size"] / values["total_steps"]
                )
            except ZeroDivisionError:
                self.macro_inputs["step_size"].set_value(units.Quantity(0, self.attribute[UNIT]))
        else:
            try:
                self.macro_inputs["total_steps"].set_value(
                    units.Quantity(math.ceil(values["field_size"] / values["step_size"]))
                )
            except ZeroDivisionError:
                self.macro_inputs["total_steps"].set_value(units.Quantity(0))

        self.macro_inputs["step_size"].input_widget.blockSignals(False)
        self.macro_inputs["total_steps"].input_widget.blockSignals(False)

    def signal_field_size_check_limits(self):
        values = self.get_input_values()
        if values["field_size"] < values["step_size"]:
            self.macro_inputs["field_size"].set_value(values["step_size"])

    def signal_step_size_changed(self, __):
        values = self.get_input_values()
        self.macro_inputs["total_steps"].input_widget.blockSignals(True)
        self.macro_inputs["field_size"].input_widget.blockSignals(True)
        if self.field_size_lock_button.isChecked():
            try:
                self.macro_inputs["total_steps"].set_value(
                    units.Quantity(math.ceil(values["field_size"] / values["step_size"]))
                )
            except ZeroDivisionError:
                self.macro_inputs["total_steps"].set_value(units.Quantity(0))
        else:
            self.macro_inputs["field_size"].set_value(
                values["step_size"] * values["total_steps"]
            )
        self.macro_inputs["total_steps"].input_widget.blockSignals(False)
        self.macro_inputs["field_size"].input_widget.blockSignals(False)

    def signal_step_size_check_limits(self):
        values = self.get_input_values()
        if values["step_size"] > values["field_size"]:
            self.macro_inputs["step_size"].set_value(values["field_size"])
        elif (values["total_steps"] * values["step_size"]) > self.macro_inputs["field_size"].get_maximum():
            self.macro_inputs["step_size"].set_value(
                self.macro_inputs["field_size"].get_maximum() / values["total_steps"]
            )

    def signal_total_steps_changed(self, __):
        values = self.get_input_values()
        self.macro_inputs["step_size"].input_widget.blockSignals(True)
        self.macro_inputs["field_size"].input_widget.blockSignals(True)
        if self.step_size_lock_button.isChecked():
            self.macro_inputs["field_size"].set_value(
                values["total_steps"] * values["step_size"]
            )
        else:
            try:
                self.macro_inputs["step_size"].set_value(
                    values["field_size"] / values["total_steps"]
                )
            except ZeroDivisionError:
                self.macro_inputs["step_size"].set_value(units.Quantity(0, self.attribute[UNIT]))
        self.macro_inputs["step_size"].input_widget.blockSignals(False)
        self.macro_inputs["field_size"].input_widget.blockSignals(False)

    def signal_total_steps_check_limits(self):
        values = self.get_input_values()
        if values["field_size"] < values["step_size"]:
            self.macro_inputs["total_steps"].set_value(
                units.Quantity(1, values["step_size"].u)
            )
        if (values["total_steps"] * values["step_size"]) > self.macro_inputs["field_size"].get_maximum():
            self.macro_inputs["total_steps"].set_value(
                units.Quantity(math.ceil(self.macro_inputs["field_size"].get_maximum() / values["step_size"]))
            )

    def set_relative_scan(self, flag):
        if flag:
            if self.config.get("attribute") is not None:
                self.attribute_selected(self.config.get("attribute"))
            else:
                self.attribute_selected(self.combobox_attribute.currentText())

    def set_absolute_scan(self, flag):
        if flag:
            if self.config.get("attribute") is not None:
                self.attribute_selected(self.config.get("attribute"))
            else:
                self.attribute_selected(self.combobox_attribute.currentText())

    def save(self):
        values = self.get_input_values()
        try:
            if self.radio_relative_scan.isChecked():
                half_field_size = values["field_size"] / 2
                values["start_point"] = self.attribute.value() - half_field_size
                values["end_point"] = self.attribute.value() + half_field_size
            step_attributes = {
                "device_id": self.combobox_device.currentText(),
                "wait_after": values["wait_after"],
                "attribute": self.config.get("attribute").split(u" # "),
                "negative_tolerance": values["positive_tolerance"],
                "positive_tolerance": values["negative_tolerance"],
                "retry_count": int(values["retry_count"].m),
                "trigger_log": self.config.get("trigger_log", True)
            }
            if values["total_steps"] <= 0:
                raise ValueError("Total steps for {} has to be greater than zero.".format(self.title))

            scanner = self.config.get("scanner", "None")
            if scanner == "None":
                scanner = None
            scanner_attributes = {}
            if scanner is not None:
                scanner_device = kamzik3.session.get_device(scanner)
                for attribute in scanner_device.get_scanner_attributes():
                    scanner_attributes[attribute] = self.macro_inputs[attribute].get_attribute_value()
            if self.config.get("as_step", False):
                step_attributes["repeat_count"] = values["repeat"]
                scan = Scan("Scan", values["start_point"], values["end_point"], int(values["total_steps"].m),
                            step_class=StepDeviceAttributeNumerical, step_attributes=step_attributes, as_step=True,
                            bidirectional=self.config.get("bidirectional", False), scanner=scanner,
                            scanner_attributes=scanner_attributes, return_back=self.config.get("return_back", True))
            else:
                scan = Scan("Scan", values["start_point"], values["end_point"], int(values["total_steps"].m),
                            repeat_count=values["repeat"], step_class=StepDeviceAttributeNumerical,
                            step_attributes=step_attributes, as_step=False,
                            bidirectional=self.config.get("bidirectional", False), scanner=scanner,
                            scanner_attributes=scanner_attributes, return_back=self.config.get("return_back", True))
            return scan
        except KeyError:
            if self.combobox_device.currentText() == "None":
                raise KeyError("Device was not set for {}".format(self.title))
            else:
                raise KeyError("One of attributes for {} were not set".format(self.title))

    def get_template(self):
        if self.combobox_device.currentText() == "None":
            raise ValueError("Attribute cannot be set to None")
        if self.combobox_attribute.currentText() == "None":
            raise ValueError("Attribute cannot be set to None")

        return {
            "type": "iterative",
            "input": "relative" if self.radio_relative_scan.isChecked() else "absolute",
            "device_id": self.combobox_device.currentText(),
            "attribute": self.combobox_attribute.currentText(),
            "trigger_log": self.checkbox_log.isChecked(),
            "as_step": self.checkbox_as_step.isChecked(),
            "scanner": self.combobox_scanner.currentText(),
            "defaults": self.checkbox_defaults.isChecked(),
            "bidirectional": self.checkbox_bidirectional.isChecked(),
            "return_back": self.checkbox_return_back.isChecked()
        }

    def show_debug_widget(self):
        if self.device is None:
            show_error_message("No device selected", parent=self)
            return
        if self.debug_widget is None:
            self.debug_widget = DeviceDebugWidget(self.device)
        elif self.debug_widget.device != self.device:
            self.debug_widget.close()
            self.debug_widget.setParent(None)
            self.debug_widget.deleteLater()
            self.debug_widget = None
            self.debug_widget = DeviceDebugWidget(self.device)

        show_centralized_top(self.debug_widget)



import logging
import os
import re
import time
from threading import Thread

import numpy as np

import kamzik3
from kamzik3 import DeviceError, units, MacroException
from kamzik3.constants import *
from kamzik3.devices.attributeLogger import AttributeLoggerTriggered
from kamzik3.devices.device import Device
from kamzik3.macro.common import Common
from kamzik3.macro.common import StepGenerator
from kamzik3.macro.scan import Scan
from kamzik3.macro.step import Step
from kamzik3.snippets.snippetDataAccess import get_next_file_index
from kamzik3.snippets.snippetsDecorators import expose_method
from kamzik3.snippets.snippetsUnits import seconds_to_datetime, get_scaled_time_duration


class Macro(Common, Device):
    current_point_index = 0
    template_id = None

    def __init__(self, device_id, chain=None, repeat_count=0, wait_after=units.Quantity(0, "s"), retry_count=0,
                 trigger_log=True, comment=None):
        self.logger = logging.getLogger("Macro.{}".format(device_id))
        self.comment = comment
        self.chain = chain
        Common.__init__(self, device_id, repeat_count, wait_after, retry_count, trigger_log)
        Device.__init__(self, device_id)
        self.macro_logger = None
        if self.chain is None:
            self.chain = []
        self.branch = []
        self.index = -1
        self.macro_body_thread = None
        self.connect()

    def _init_attributes(self):
        super(Macro, self)._init_attributes()
        self.create_attribute(ATTR_STATE, default_vlaue=IDLE, readonly=True, default_type=np.uint8)
        self.create_attribute(ATTR_POINTS_COUNT, default_vlaue=1, readonly=True, default_type=np.uint64)
        self.create_attribute(ATTR_POINT_INDEX, default_vlaue=0, readonly=True, default_type=np.uint64)
        self.create_attribute(ATTR_PROGRESS, default_vlaue=0, readonly=True, default_type=np.float16, unit="%",
                              min_value=0, max_value=100, decimals=2)
        self.create_attribute(ATTR_CHAIN_LINK, default_vlaue=0, readonly=True, default_type=np.uint64)
        self.create_attribute(ATTR_CHAIN_LINK_OUTPUT, default_vlaue="", readonly=True)
        self.create_attribute(ATTR_RUNNING_TIME, default_vlaue=0, readonly=True)
        self.create_attribute(ATTR_ESTIMATED_TIME, default_vlaue=0, readonly=True)
        self.create_attribute(ATTR_ESTIMATED_END, default_vlaue=0, readonly=True)
        self.create_attribute(ATTR_REMAINING_TIME, default_vlaue=0, readonly=True)
        self.create_attribute(ATTR_STARTED_AT, default_vlaue=None, readonly=True)
        self.create_attribute(ATTR_OUTPUT_FILENAME, default_vlaue=None, readonly=True)
        self.set_value(ATTR_DESCRIPTION, self.comment)

    def handle_configuration(self):
        start_at = time.time()

        def _finish_configuration(*_, **__):
            self._config_commands()
            self._config_attributes()
            self.set_status(STATUS_CONFIGURED)
            self.recount_steps()
            self.logger.info(u"Device configuration took {} sec.".format(time.time() - start_at))

        _finish_configuration()

    def command(self, command, callback=None, with_token=False, returning=True):
        raise DeviceError("Device does not accept any commands.")

    def add(self, link):
        if isinstance(link, list):
            for link_item in link:
                self.add(link_item)
            return
        assert isinstance(link, Common)
        if isinstance(link, Macro):
            self.chain += link.chain
        else:
            self.chain.append(link)
        self.recount_steps()

    def next(self):
        self.index += 1
        if self.index < len(self.chain):
            return self.chain[self.index]
        else:
            return None

    def prev(self):
        self.index -= 1
        if self.index >= 0:
            return self.chain[self.index]
        else:
            return None

    def reset(self):
        self.index = -1
        self.set_point_index(0)
        super(Macro, self).reset()

    def resolve_scans(self):
        new_chain = []
        for link in self.chain:
            if isinstance(link, Scan) and link.scanner is not None and not link.resolved:
                scanner = kamzik3.session.get_device(link.scanner)
                scan_macro = scanner.get_scanner_macro(scanner_input=link,
                                                       scanner_attributes=link.scanner_attributes)
                new_chain += scan_macro.chain
                link.resolved = True
            else:
                new_chain.append(link)
        self.chain = new_chain
        self.recount_steps()

    def before_macro_starts(self):
        pass

    def before_macro_stops(self):
        pass

    @expose_method()
    def start(self, threaded=True):
        if self.get_state() == BUSY:
            raise DeviceError(u"Macro is already running")
        self.resolve_scans()
        super(Macro, self).start()
        if self.macro_logger is not None:
            self.macro_logger.start()

        self.set_point_index(0)
        self.set_value(ATTR_STARTED_AT, seconds_to_datetime(self.started_at))
        self.set_value(ATTR_ESTIMATED_END, seconds_to_datetime(self.started_at))
        self.current_point_index = 0
        self.index = -1
        self.before_macro_starts()
        if threaded:
            self.macro_body_thread = Thread(target=self.body, args=[self.macro_clean])
            self.macro_body_thread.start()
        else:
            self.body(callback=self.macro_clean)

    def body(self, callback=None):
        try:
            # Repeat following macro steps
            for repeat in range(self.repeat_count + 1):
                # Reset all links in chain
                for link in self.chain:
                    link.reset()
                # Get first link
                link = self.next()
                # Reset branch to empty list
                self.branch = []
                # Stay in Macro loop until macro is BUSY
                while self.get_state() == BUSY:
                    # Link is empty
                    if link is None:
                        if self.branch:
                            if self.branch[-1] in self.chain:
                                link = self.prev()
                            else:
                                link = self.branch.pop()
                                self.index -= 1
                        else:
                            link = self.prev()
                        continue
                    # Link is done and repeated
                    elif link.get_state() == DONE and link.repeated == link.repeat_count:
                        link.reset()
                        link = self.prev()
                        if link is None:
                            if self.repeated == self.repeat_count:
                                self.done()
                            else:
                                self.repeated += 1
                                break
                        continue
                    # Deal with Step type of Link
                    elif isinstance(link, Step):
                        if link.get_state() == DONE:
                            link.repeat()
                        else:
                            try:
                                link.start()
                            except MacroException as retry_reason:
                                link.retry(retry_reason)
                                continue
                    # Deal with StepGenerator type of Link
                    elif isinstance(link, StepGenerator):
                        if link.as_step:
                            step = link.start()
                            link = step
                            continue
                        # Get step from step generator
                        if link.get_state() == DONE:
                            step = link.repeat()
                        elif self.get_state() != BUSY:
                            continue
                        else:
                            try:
                                step = link.start()
                            except MacroException as retry_reason:
                                link.retry(retry_reason)
                                continue
                        if step is None:
                            if self.branch:
                                if link.repeated == link.repeat_count:
                                    link = self.branch.pop()
                            else:
                                link = self.chain[self.index]
                        else:
                            if link not in self.branch:
                                self.branch.append(link)
                            link = step
                        continue

                    # Set link index of chain
                    self.set_value(ATTR_CHAIN_LINK, self.index)
                    # Set output of link
                    self.set_attribute((ATTR_CHAIN_LINK_OUTPUT, VALUE),
                                       [self.current_point_index + 1, self.index, link.get_output()])
                    # Set point index of macro
                    self.set_point_index(self.current_point_index + 1)
                    # Write into log file if trigger_log is set and we have logger associated
                    if link.trigger_log and self.macro_logger:
                        self.macro_logger.trigger()

                    if link.repeated == link.repeat_count:
                        link = self.next()

                if self.in_statuses(IDLE_DEVICE_STATUSES):
                    return

                self.start_timer()
        except MacroException as e:
            self.error(e)
            raise e
        finally:
            if callable(callback):
                callback()

    @expose_method()
    def stop(self):
        if self.in_statuses(IDLE_DEVICE_STATUSES):
            raise DeviceError(u"Macro is not running")
        for link in self.chain + self.branch:
            link.stop()
        self.before_macro_stops()
        super(Macro, self).stop()

    def macro_clean(self):
        for link in self.chain:
            if isinstance(link, StepGenerator):
                link.return_back_callback()
        self.before_macro_stops()

    def recount_steps(self):
        sub_steps_count = 0
        sub_points_count = 0
        for link in reversed(self.chain):
            if isinstance(link, StepGenerator) and sub_steps_count > 0:
                repeat_step = (link.step_attributes.get("repeat_count", 0) + 1)
                repeat_scan = (link.repeat_count + 1)
                total_steps = link.steps_count * sub_steps_count
                sub_steps_count = total_steps * repeat_scan

                total_points = (link.steps_count + 1) * (repeat_step + sub_points_count)
                sub_points_count = total_points * repeat_scan
            else:
                sub_steps_count += link.get_total_steps_count()
                sub_points_count += link.get_total_points_count()

        total_count = sub_points_count * (self.repeat_count + 1)
        self.points_count = total_count
        self.set_value(ATTR_POINTS_COUNT, self.points_count)

    def get_total_steps_count(self):
        return len(self.chain)

    def get_total_points_count(self):
        return self.points_count

    @expose_method()
    def get_chain_link_output(self):
        chain_link_output = []
        for index, chain_item in enumerate(self.chain):
            if isinstance(chain_item, Scan):
                chain_link_output.append((chain_item.step_attributes["device_id"],
                                          chain_item.step_attributes["attribute"], chain_item.points_count))
            else:
                chain_link_output.append(None)
        return chain_link_output

    def set_state(self, new_state):
        super(Macro, self).set_state(new_state)

        self.set_value(ATTR_STATE, new_state)
        if new_state in (IDLE, DONE, STOPPED):
            self.set_status(STATUS_IDLE)
        elif new_state in (BUSY, PAUSED):
            self.set_status(STATUS_BUSY)
        elif new_state == ERROR:
            self.set_status(STATUS_ERROR)

        if self.macro_logger is not None and new_state in (DONE, ERROR, IDLE, DONE, STOPPED):
            self.macro_logger.stop()

    def set_point_index(self, point_index):
        self.current_point_index = point_index
        self.set_value(ATTR_POINT_INDEX, point_index)
        progress = (100. / self[ATTR_POINTS_COUNT][VALUE]) * point_index
        self.set_value(ATTR_PROGRESS, progress)

        running_time = time.time() - self.started_at
        try:
            estimated_time = (100 / progress) * running_time
        except ZeroDivisionError:
            estimated_time = 1
        remaining_time = estimated_time - running_time
        self.set_value(ATTR_RUNNING_TIME, get_scaled_time_duration(running_time))
        self.set_value(ATTR_REMAINING_TIME, get_scaled_time_duration(remaining_time))
        self.set_value(ATTR_ESTIMATED_END, seconds_to_datetime(time.time() + remaining_time))
        self.set_value(ATTR_ESTIMATED_TIME, get_scaled_time_duration(estimated_time))

    def setup_output_file(self, output_directory_path, macro_prefix):
        if not os.path.exists(output_directory_path):
            self.logger.info(u"Output directory {} does not exists, trying to create it.".format(output_directory_path))
            os.makedirs(output_directory_path)
        next_index = get_next_file_index(output_directory_path, re.compile("{}_(\d+)\.log".format(macro_prefix)))
        output_filename = os.path.join(output_directory_path, "{}_{}.log".format(macro_prefix, next_index))
        self.set_value(ATTR_OUTPUT_FILENAME, output_filename)
        self.macro_logger = AttributeLoggerTriggered(output_filename,
                                                     device_id="{}_Log_{}".format(macro_prefix, next_index),
                                                     config={"rotating": False})
        return output_filename, next_index

    def add_output_attribute(self, device_id, attribute):
        if self.macro_logger is None:
            raise DeviceError(u"Macro logger is not set")

        self.macro_logger.add_logged_attribute(device_id, attribute)

    def get_output_header(self):
        lines = []
        for link in self.chain:
            lines += link.get_output_header()
            lines += ["-" * 32]
        return lines

    def disconnect(self):
        if self.chain is not None:
            for chain in self.chain:
                chain.remove()

        if self.macro_logger is not None:
            self.macro_logger.remove()

        return Device.disconnect(self)

    def remove(self):
        for step in self.chain:
            step.remove()
        Device.remove(self)
        self.macro_logger = None
        self.chain = None
        Common.remove(self)

    def error(self, error_message):
        self.error_at = time.time()
        self.running_time = self.error_at - self.started_at
        self.error_message = error_message
        self.set_value(ATTR_LAST_ERROR, error_message)
        self.logger.error(error_message)
        self.handle_command_error(self.common_id, error_message)
        self.set_state(ERROR)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'macroToolTemplate.ui'
#
# Created by: PyQt5 UI code generator 5.13.0
#
# WARNING! All changes made in this file will be lost!


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(803, 883)
        self.gridLayout_2 = QtWidgets.QGridLayout(Form)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.splitter = QtWidgets.QSplitter(Form)
        self.splitter.setOrientation(QtCore.Qt.Horizontal)
        self.splitter.setObjectName("splitter")
        self.widget_macro_designer = QtWidgets.QWidget(self.splitter)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.widget_macro_designer.sizePolicy().hasHeightForWidth())
        self.widget_macro_designer.setSizePolicy(sizePolicy)
        self.widget_macro_designer.setMaximumSize(QtCore.QSize(450, 16777215))
        self.widget_macro_designer.setObjectName("widget_macro_designer")
        self.gridLayout = QtWidgets.QGridLayout(self.widget_macro_designer)
        self.gridLayout.setContentsMargins(0, 0, 0, 0)
        self.gridLayout.setObjectName("gridLayout")
        self.layout_template_select = QtWidgets.QFormLayout()
        self.layout_template_select.setFieldGrowthPolicy(QtWidgets.QFormLayout.AllNonFixedFieldsGrow)
        self.layout_template_select.setLabelAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.layout_template_select.setObjectName("layout_template_select")
        self.label_template = QtWidgets.QLabel(self.widget_macro_designer)
        self.label_template.setObjectName("label_template")
        self.layout_template_select.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.label_template)
        self.combobox_template = QtWidgets.QComboBox(self.widget_macro_designer)
        self.combobox_template.setObjectName("combobox_template")
        self.combobox_template.addItem("")
        self.layout_template_select.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.combobox_template)
        self.gridLayout.addLayout(self.layout_template_select, 1, 0, 1, 1)
        self.label_new_macro = QtWidgets.QLabel(self.widget_macro_designer)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_new_macro.sizePolicy().hasHeightForWidth())
        self.label_new_macro.setSizePolicy(sizePolicy)
        self.label_new_macro.setStyleSheet("font-weight:bold;background:#f6b442;padding:5")
        self.label_new_macro.setObjectName("label_new_macro")
        self.gridLayout.addWidget(self.label_new_macro, 0, 0, 1, 1)
        self.widget_macro_components = QtWidgets.QStackedWidget(self.widget_macro_designer)
        self.widget_macro_components.setObjectName("widget_macro_components")
        self.page_macro_components = QtWidgets.QWidget()
        self.page_macro_components.setObjectName("page_macro_components")
        self.gridLayout_4 = QtWidgets.QGridLayout(self.page_macro_components)
        self.gridLayout_4.setContentsMargins(0, 0, 0, 0)
        self.gridLayout_4.setObjectName("gridLayout_4")
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.button_start_macro = QtWidgets.QPushButton(self.page_macro_components)
        self.button_start_macro.setStatusTip("")
        self.button_start_macro.setStyleSheet("QPushButton { background-color: #68d532 }\\nQPushButton:pressed { background-color: #28a900 }\\nQPushButton:!enabled{background:silver;}QPushButton:checked{background:red;}\n"
"QPushButton:disabled{ background:silver;color:grey }")
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/icons/icons/play.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.button_start_macro.setIcon(icon)
        self.button_start_macro.setIconSize(QtCore.QSize(16, 16))
        self.button_start_macro.setObjectName("button_start_macro")
        self.horizontalLayout.addWidget(self.button_start_macro)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        self.button_remove = QtWidgets.QPushButton(self.page_macro_components)
        self.button_remove.setStyleSheet("QPushButton { background-color: #ff4e00 }\\\\nQPushButton:pressed { background-color: #ff4e00 }\\\\nQPushButton:!enabled{background:silver;}QPushButton:checked{background:red;}\\nQPushButton:disabled{ background:silver;color:grey }")
        icon1 = QtGui.QIcon()
        icon1.addPixmap(QtGui.QPixmap(":/icons/icons/remove.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.button_remove.setIcon(icon1)
        self.button_remove.setIconSize(QtCore.QSize(16, 16))
        self.button_remove.setObjectName("button_remove")
        self.horizontalLayout.addWidget(self.button_remove)
        self.gridLayout_4.addLayout(self.horizontalLayout, 3, 0, 1, 1)
        self.formLayout = QtWidgets.QFormLayout()
        self.formLayout.setFieldGrowthPolicy(QtWidgets.QFormLayout.AllNonFixedFieldsGrow)
        self.formLayout.setLabelAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.formLayout.setObjectName("formLayout")
        self.label_2 = QtWidgets.QLabel(self.page_macro_components)
        self.label_2.setObjectName("label_2")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.label_2)
        self.input_comment = QtWidgets.QLineEdit(self.page_macro_components)
        self.input_comment.setObjectName("input_comment")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.input_comment)
        self.label_4 = QtWidgets.QLabel(self.page_macro_components)
        self.label_4.setObjectName("label_4")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.LabelRole, self.label_4)
        self.input_repeat_macro = QtWidgets.QDoubleSpinBox(self.page_macro_components)
        self.input_repeat_macro.setDecimals(0)
        self.input_repeat_macro.setMaximum(2140000000.0)
        self.input_repeat_macro.setObjectName("input_repeat_macro")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.FieldRole, self.input_repeat_macro)
        self.gridLayout_4.addLayout(self.formLayout, 1, 0, 1, 1)
        self.scroll_area_macro = QtWidgets.QScrollArea(self.page_macro_components)
        self.scroll_area_macro.setWidgetResizable(True)
        self.scroll_area_macro.setObjectName("scroll_area_macro")
        self.widget_macro_list = QtWidgets.QWidget()
        self.widget_macro_list.setGeometry(QtCore.QRect(0, 0, 448, 383))
        self.widget_macro_list.setObjectName("widget_macro_list")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.widget_macro_list)
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout.setObjectName("verticalLayout")
        self.scroll_area_macro.setWidget(self.widget_macro_list)
        self.gridLayout_4.addWidget(self.scroll_area_macro, 2, 0, 1, 1)
        self.widget_macro_components.addWidget(self.page_macro_components)
        self.page_template_designer = QtWidgets.QWidget()
        self.page_template_designer.setObjectName("page_template_designer")
        self.gridLayout_5 = QtWidgets.QGridLayout(self.page_template_designer)
        self.gridLayout_5.setContentsMargins(0, 0, 0, 0)
        self.gridLayout_5.setObjectName("gridLayout_5")
        self.scroll_area_custom_template = QtWidgets.QScrollArea(self.page_template_designer)
        self.scroll_area_custom_template.setWidgetResizable(True)
        self.scroll_area_custom_template.setObjectName("scroll_area_custom_template")
        self.widget_custom_template = QtWidgets.QWidget()
        self.widget_custom_template.setGeometry(QtCore.QRect(0, 0, 98, 28))
        self.widget_custom_template.setObjectName("widget_custom_template")
        self.verticalLayout_5 = QtWidgets.QVBoxLayout(self.widget_custom_template)
        self.verticalLayout_5.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_5.setObjectName("verticalLayout_5")
        self.scroll_area_custom_template.setWidget(self.widget_custom_template)
        self.gridLayout_5.addWidget(self.scroll_area_custom_template, 0, 0, 1, 1)
        self.horizontalLayout_5 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_5.setObjectName("horizontalLayout_5")
        self.button_reset_template = QtWidgets.QPushButton(self.page_template_designer)
        self.button_reset_template.setStatusTip("")
        self.button_reset_template.setStyleSheet("QPushButton { background-color: #ff4e00 }\\\\nQPushButton:pressed { background-color: #ff4e00 }\\\\nQPushButton:!enabled{background:silver;}QPushButton:checked{background:red;}\\nQPushButton:disabled{ background:silver;color:grey }")
        icon2 = QtGui.QIcon()
        icon2.addPixmap(QtGui.QPixmap(":/icons/icons/reset_bw.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.button_reset_template.setIcon(icon2)
        self.button_reset_template.setIconSize(QtCore.QSize(16, 16))
        self.button_reset_template.setObjectName("button_reset_template")
        self.horizontalLayout_5.addWidget(self.button_reset_template)
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_5.addItem(spacerItem1)
        self.button_save_template = QtWidgets.QPushButton(self.page_template_designer)
        self.button_save_template.setStyleSheet("QPushButton { background-color: #00b4ff }\\nQPushButton:pressed { background-color: #28a900 }\\nQPushButton:!enabled{background:silver;}QPushButton:checked{background:red;}\n"
"QPushButton:disabled{ background:silver;color:grey }")
        icon3 = QtGui.QIcon()
        icon3.addPixmap(QtGui.QPixmap(":/icons/icons/save_bw.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.button_save_template.setIcon(icon3)
        self.button_save_template.setIconSize(QtCore.QSize(16, 16))
        self.button_save_template.setObjectName("button_save_template")
        self.horizontalLayout_5.addWidget(self.button_save_template)
        self.gridLayout_5.addLayout(self.horizontalLayout_5, 5, 0, 1, 1)
        self.horizontalLayout_4 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_4.setObjectName("horizontalLayout_4")
        self.button_add_iterative_step = QtWidgets.QPushButton(self.page_template_designer)
        self.button_add_iterative_step.setStatusTip("")
        self.button_add_iterative_step.setStyleSheet("QPushButton { background-color: #68d532 }\\nQPushButton:pressed { background-color: #28a900 }\\nQPushButton:!enabled{background:silver;}QPushButton:checked{background:red;}\n"
"QPushButton:disabled{ background:silver;color:grey }")
        self.button_add_iterative_step.setIconSize(QtCore.QSize(16, 16))
        self.button_add_iterative_step.setObjectName("button_add_iterative_step")
        self.horizontalLayout_4.addWidget(self.button_add_iterative_step)
        self.button_add_action_step = QtWidgets.QPushButton(self.page_template_designer)
        self.button_add_action_step.setStatusTip("")
        self.button_add_action_step.setStyleSheet("QPushButton { background-color: #68d532 }\\nQPushButton:pressed { background-color: #28a900 }\\nQPushButton:!enabled{background:silver;}QPushButton:checked{background:red;}\n"
"QPushButton:disabled{ background:silver;color:grey }")
        self.button_add_action_step.setIconSize(QtCore.QSize(16, 16))
        self.button_add_action_step.setObjectName("button_add_action_step")
        self.horizontalLayout_4.addWidget(self.button_add_action_step)
        spacerItem2 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_4.addItem(spacerItem2)
        self.gridLayout_5.addLayout(self.horizontalLayout_4, 2, 0, 1, 1)
        self.label_measurement_group = QtWidgets.QLabel(self.page_template_designer)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_measurement_group.sizePolicy().hasHeightForWidth())
        self.label_measurement_group.setSizePolicy(sizePolicy)
        self.label_measurement_group.setStyleSheet("font-weight:bold;background:#f6b442;padding:5")
        self.label_measurement_group.setObjectName("label_measurement_group")
        self.gridLayout_5.addWidget(self.label_measurement_group, 3, 0, 1, 1)
        self.list_measurement_groups = QtWidgets.QListWidget(self.page_template_designer)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.list_measurement_groups.sizePolicy().hasHeightForWidth())
        self.list_measurement_groups.setSizePolicy(sizePolicy)
        self.list_measurement_groups.setSelectionMode(QtWidgets.QAbstractItemView.MultiSelection)
        self.list_measurement_groups.setObjectName("list_measurement_groups")
        self.gridLayout_5.addWidget(self.list_measurement_groups, 4, 0, 1, 1)
        self.widget_macro_components.addWidget(self.page_template_designer)
        self.gridLayout.addWidget(self.widget_macro_components, 3, 0, 1, 1)
        self.tab_widget_macro_list = QtWidgets.QTabWidget(self.splitter)
        self.tab_widget_macro_list.setAutoFillBackground(False)
        self.tab_widget_macro_list.setStyleSheet("")
        self.tab_widget_macro_list.setDocumentMode(True)
        self.tab_widget_macro_list.setTabsClosable(True)
        self.tab_widget_macro_list.setObjectName("tab_widget_macro_list")
        self.gridLayout_2.addWidget(self.splitter, 0, 0, 1, 1)

        self.retranslateUi(Form)
        self.widget_macro_components.setCurrentIndex(0)
        self.combobox_template.currentTextChanged['QString'].connect(Form.template_selected)
        self.button_add_iterative_step.clicked.connect(Form.add_iterative_step)
        self.button_add_action_step.clicked.connect(Form.add_action_step)
        self.button_reset_template.clicked.connect(Form.reset_template_input)
        self.button_save_template.clicked.connect(Form.export_template)
        self.button_remove.clicked.connect(Form.remove_template)
        self.button_start_macro.clicked.connect(Form.start_macro)
        self.tab_widget_macro_list.tabCloseRequested['int'].connect(Form.close_macro_tab)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Scanning tool"))
        self.label_template.setText(_translate("Form", "Template:"))
        self.combobox_template.setItemText(0, _translate("Form", "Custom"))
        self.label_new_macro.setText(_translate("Form", "Create new macro"))
        self.button_start_macro.setToolTip(_translate("Form", "Start scan"))
        self.button_start_macro.setText(_translate("Form", "Start"))
        self.button_remove.setToolTip(_translate("Form", "Remove selected template"))
        self.button_remove.setText(_translate("Form", "Remove"))
        self.label_2.setText(_translate("Form", "Comment:"))
        self.label_4.setText(_translate("Form", "Repeat:"))
        self.button_reset_template.setToolTip(_translate("Form", "Reset template"))
        self.button_reset_template.setText(_translate("Form", "Reset"))
        self.button_save_template.setToolTip(_translate("Form", "Save template"))
        self.button_save_template.setText(_translate("Form", "Save"))
        self.button_add_iterative_step.setToolTip(_translate("Form", "Add iterative step"))
        self.button_add_iterative_step.setText(_translate("Form", "+ Iterative step"))
        self.button_add_action_step.setToolTip(_translate("Form", "Add action step"))
        self.button_add_action_step.setText(_translate("Form", "+ Action step"))
        self.label_measurement_group.setText(_translate("Form", "Select measurement group(s)"))
        self.list_measurement_groups.setSortingEnabled(True)
from kamzik3 import resource_kamzik3_rc



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'savedAttributesTemplate.ui'
#
# Created by: PyQt5 UI code generator 5.10.1
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(974, 644)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(Form.sizePolicy().hasHeightForWidth())
        Form.setSizePolicy(sizePolicy)
        self.gridLayout = QtWidgets.QGridLayout(Form)
        self.gridLayout.setObjectName("gridLayout")
        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.verticalLayout.setObjectName("verticalLayout")
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout.addItem(spacerItem)
        self.button_set_value = QtWidgets.QPushButton(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.button_set_value.sizePolicy().hasHeightForWidth())
        self.button_set_value.setSizePolicy(sizePolicy)
        self.button_set_value.setMinimumSize(QtCore.QSize(35, 35))
        self.button_set_value.setMaximumSize(QtCore.QSize(40, 40))
        self.button_set_value.setText("")
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/icons/icons/target.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.button_set_value.setIcon(icon)
        self.button_set_value.setIconSize(QtCore.QSize(25, 25))
        self.button_set_value.setObjectName("button_set_value")
        self.verticalLayout.addWidget(self.button_set_value)
        spacerItem1 = QtWidgets.QSpacerItem(20, 27, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        self.verticalLayout.addItem(spacerItem1)
        self.button_change_color = QtWidgets.QPushButton(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.button_change_color.sizePolicy().hasHeightForWidth())
        self.button_change_color.setSizePolicy(sizePolicy)
        self.button_change_color.setMinimumSize(QtCore.QSize(35, 35))
        self.button_change_color.setMaximumSize(QtCore.QSize(40, 40))
        font = QtGui.QFont()
        font.setPointSize(11)
        self.button_change_color.setFont(font)
        self.button_change_color.setStyleSheet("QPushButton {color:red}")
        self.button_change_color.setText("")
        icon1 = QtGui.QIcon()
        icon1.addPixmap(QtGui.QPixmap(":/icons/icons/color_picker.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.button_change_color.setIcon(icon1)
        self.button_change_color.setIconSize(QtCore.QSize(25, 25))
        self.button_change_color.setObjectName("button_change_color")
        self.verticalLayout.addWidget(self.button_change_color)
        self.button_move_top = QtWidgets.QPushButton(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.button_move_top.sizePolicy().hasHeightForWidth())
        self.button_move_top.setSizePolicy(sizePolicy)
        self.button_move_top.setMinimumSize(QtCore.QSize(35, 35))
        self.button_move_top.setMaximumSize(QtCore.QSize(40, 40))
        self.button_move_top.setText("")
        icon2 = QtGui.QIcon()
        icon2.addPixmap(QtGui.QPixmap(":/icons/icons/top.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.button_move_top.setIcon(icon2)
        self.button_move_top.setIconSize(QtCore.QSize(25, 25))
        self.button_move_top.setObjectName("button_move_top")
        self.verticalLayout.addWidget(self.button_move_top)
        self.button_move_up = QtWidgets.QPushButton(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.button_move_up.sizePolicy().hasHeightForWidth())
        self.button_move_up.setSizePolicy(sizePolicy)
        self.button_move_up.setMinimumSize(QtCore.QSize(35, 35))
        self.button_move_up.setMaximumSize(QtCore.QSize(40, 40))
        self.button_move_up.setText("")
        icon3 = QtGui.QIcon()
        icon3.addPixmap(QtGui.QPixmap(":/icons/icons/plus.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.button_move_up.setIcon(icon3)
        self.button_move_up.setIconSize(QtCore.QSize(25, 25))
        self.button_move_up.setCheckable(False)
        self.button_move_up.setChecked(False)
        self.button_move_up.setObjectName("button_move_up")
        self.verticalLayout.addWidget(self.button_move_up)
        spacerItem2 = QtWidgets.QSpacerItem(20, 20, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        self.verticalLayout.addItem(spacerItem2)
        self.button_remove = QtWidgets.QPushButton(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.button_remove.sizePolicy().hasHeightForWidth())
        self.button_remove.setSizePolicy(sizePolicy)
        self.button_remove.setMinimumSize(QtCore.QSize(35, 35))
        self.button_remove.setMaximumSize(QtCore.QSize(40, 40))
        font = QtGui.QFont()
        font.setPointSize(15)
        font.setBold(True)
        font.setWeight(75)
        self.button_remove.setFont(font)
        self.button_remove.setStyleSheet("QPushButton {color:red}")
        self.button_remove.setIconSize(QtCore.QSize(0, 0))
        self.button_remove.setObjectName("button_remove")
        self.verticalLayout.addWidget(self.button_remove)
        self.button_add = QtWidgets.QPushButton(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.button_add.sizePolicy().hasHeightForWidth())
        self.button_add.setSizePolicy(sizePolicy)
        self.button_add.setMinimumSize(QtCore.QSize(35, 35))
        self.button_add.setMaximumSize(QtCore.QSize(40, 40))
        font = QtGui.QFont()
        font.setPointSize(15)
        font.setBold(True)
        font.setWeight(75)
        self.button_add.setFont(font)
        self.button_add.setStyleSheet("QPushButton {color:green}")
        self.button_add.setIconSize(QtCore.QSize(0, 0))
        self.button_add.setObjectName("button_add")
        self.verticalLayout.addWidget(self.button_add)
        spacerItem3 = QtWidgets.QSpacerItem(20, 20, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        self.verticalLayout.addItem(spacerItem3)
        self.button_move_down = QtWidgets.QPushButton(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.button_move_down.sizePolicy().hasHeightForWidth())
        self.button_move_down.setSizePolicy(sizePolicy)
        self.button_move_down.setMinimumSize(QtCore.QSize(35, 35))
        self.button_move_down.setMaximumSize(QtCore.QSize(40, 40))
        self.button_move_down.setText("")
        icon4 = QtGui.QIcon()
        icon4.addPixmap(QtGui.QPixmap(":/icons/icons/minus.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.button_move_down.setIcon(icon4)
        self.button_move_down.setIconSize(QtCore.QSize(25, 25))
        self.button_move_down.setObjectName("button_move_down")
        self.verticalLayout.addWidget(self.button_move_down)
        self.button_move_bottom = QtWidgets.QPushButton(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.button_move_bottom.sizePolicy().hasHeightForWidth())
        self.button_move_bottom.setSizePolicy(sizePolicy)
        self.button_move_bottom.setMinimumSize(QtCore.QSize(35, 35))
        self.button_move_bottom.setMaximumSize(QtCore.QSize(40, 40))
        self.button_move_bottom.setText("")
        icon5 = QtGui.QIcon()
        icon5.addPixmap(QtGui.QPixmap(":/icons/icons/bottom.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.button_move_bottom.setIcon(icon5)
        self.button_move_bottom.setIconSize(QtCore.QSize(25, 25))
        self.button_move_bottom.setObjectName("button_move_bottom")
        self.verticalLayout.addWidget(self.button_move_bottom)
        spacerItem4 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout.addItem(spacerItem4)
        self.gridLayout.addLayout(self.verticalLayout, 1, 2, 1, 1)
        self.tab_groups = QtWidgets.QTabWidget(Form)
        self.tab_groups.setObjectName("tab_groups")
        self.gridLayout.addWidget(self.tab_groups, 1, 0, 1, 1)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.button_reset_all = QtWidgets.QPushButton(Form)
        self.button_reset_all.setToolTip("")
        self.button_reset_all.setStyleSheet("QPushButton { background-color: #ff4e00 }\\\\nQPushButton:pressed { background-color: #ff4e00 }\\\\nQPushButton:!enabled{background:silver;}QPushButton:checked{background:red;}\\nQPushButton:disabled{ background:silver;color:grey }")
        icon6 = QtGui.QIcon()
        icon6.addPixmap(QtGui.QPixmap(":/icons/icons/reset_bw.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.button_reset_all.setIcon(icon6)
        self.button_reset_all.setIconSize(QtCore.QSize(16, 16))
        self.button_reset_all.setObjectName("button_reset_all")
        self.horizontalLayout.addWidget(self.button_reset_all)
        self.button_load = QtWidgets.QPushButton(Form)
        self.button_load.setToolTip("")
        self.button_load.setStyleSheet("QPushButton { background-color: #00b4ff }\\nQPushButton:pressed { background-color: #28a900 }\\nQPushButton:!enabled{background:silver;}QPushButton:checked{background:red;}\n"
"QPushButton:disabled{ background:silver;color:grey }")
        icon7 = QtGui.QIcon()
        icon7.addPixmap(QtGui.QPixmap(":/icons/icons/load_bw.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.button_load.setIcon(icon7)
        self.button_load.setIconSize(QtCore.QSize(16, 16))
        self.button_load.setObjectName("button_load")
        self.horizontalLayout.addWidget(self.button_load)
        self.checkbox_allow_double_click = QtWidgets.QCheckBox(Form)
        self.checkbox_allow_double_click.setObjectName("checkbox_allow_double_click")
        self.horizontalLayout.addWidget(self.checkbox_allow_double_click)
        spacerItem5 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem5)
        self.checkbox_synced_with_server = QtWidgets.QCheckBox(Form)
        self.checkbox_synced_with_server.setEnabled(False)
        self.checkbox_synced_with_server.setCheckable(True)
        self.checkbox_synced_with_server.setChecked(False)
        self.checkbox_synced_with_server.setObjectName("checkbox_synced_with_server")
        self.horizontalLayout.addWidget(self.checkbox_synced_with_server)
        self.button_save = QtWidgets.QPushButton(Form)
        self.button_save.setToolTip("")
        self.button_save.setStyleSheet("QPushButton { background-color: #00b4ff }\\nQPushButton:pressed { background-color: #28a900 }\\nQPushButton:!enabled{background:silver;}QPushButton:checked{background:red;}\n"
"QPushButton:disabled{ background:silver;color:grey }")
        icon8 = QtGui.QIcon()
        icon8.addPixmap(QtGui.QPixmap(":/icons/icons/save_bw.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.button_save.setIcon(icon8)
        self.button_save.setIconSize(QtCore.QSize(16, 16))
        self.button_save.setObjectName("button_save")
        self.horizontalLayout.addWidget(self.button_save)
        self.gridLayout.addLayout(self.horizontalLayout, 2, 0, 1, 1)
        self.groups_frame = QtWidgets.QFrame(Form)
        self.groups_frame.setStyleSheet("QFrame#groups_frame {background-color:#f6b442}")
        self.groups_frame.setObjectName("groups_frame")
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout(self.groups_frame)
        self.horizontalLayout_3.setContentsMargins(6, 2, 6, 2)
        self.horizontalLayout_3.setSpacing(3)
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.label_saved_groups = QtWidgets.QLabel(self.groups_frame)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_saved_groups.sizePolicy().hasHeightForWidth())
        self.label_saved_groups.setSizePolicy(sizePolicy)
        font = QtGui.QFont()
        font.setPointSize(8)
        font.setBold(True)
        font.setWeight(75)
        self.label_saved_groups.setFont(font)
        self.label_saved_groups.setObjectName("label_saved_groups")
        self.horizontalLayout_3.addWidget(self.label_saved_groups)
        self.combo_preset = QtWidgets.QComboBox(self.groups_frame)
        self.combo_preset.setMinimumSize(QtCore.QSize(200, 0))
        self.combo_preset.setObjectName("combo_preset")
        self.horizontalLayout_3.addWidget(self.combo_preset)
        self.button_remove_preset_group = QtWidgets.QPushButton(self.groups_frame)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.button_remove_preset_group.sizePolicy().hasHeightForWidth())
        self.button_remove_preset_group.setSizePolicy(sizePolicy)
        self.button_remove_preset_group.setMinimumSize(QtCore.QSize(25, 0))
        self.button_remove_preset_group.setMaximumSize(QtCore.QSize(25, 16777215))
        font = QtGui.QFont()
        font.setPointSize(9)
        font.setBold(True)
        font.setWeight(75)
        self.button_remove_preset_group.setFont(font)
        self.button_remove_preset_group.setStyleSheet("QPushButton {color:red}")
        self.button_remove_preset_group.setIconSize(QtCore.QSize(0, 0))
        self.button_remove_preset_group.setObjectName("button_remove_preset_group")
        self.horizontalLayout_3.addWidget(self.button_remove_preset_group)
        self.button_add_preset_group = QtWidgets.QPushButton(self.groups_frame)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.button_add_preset_group.sizePolicy().hasHeightForWidth())
        self.button_add_preset_group.setSizePolicy(sizePolicy)
        self.button_add_preset_group.setMinimumSize(QtCore.QSize(25, 0))
        self.button_add_preset_group.setMaximumSize(QtCore.QSize(69, 16777215))
        font = QtGui.QFont()
        font.setPointSize(8)
        font.setBold(True)
        font.setWeight(75)
        self.button_add_preset_group.setFont(font)
        self.button_add_preset_group.setStyleSheet("QPushButton {color:green}")
        self.button_add_preset_group.setIconSize(QtCore.QSize(0, 0))
        self.button_add_preset_group.setObjectName("button_add_preset_group")
        self.horizontalLayout_3.addWidget(self.button_add_preset_group)
        self.button_add_preset_group_from_selected_rows = QtWidgets.QPushButton(self.groups_frame)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.button_add_preset_group_from_selected_rows.sizePolicy().hasHeightForWidth())
        self.button_add_preset_group_from_selected_rows.setSizePolicy(sizePolicy)
        self.button_add_preset_group_from_selected_rows.setMinimumSize(QtCore.QSize(25, 0))
        self.button_add_preset_group_from_selected_rows.setMaximumSize(QtCore.QSize(170, 16777215))
        font = QtGui.QFont()
        font.setPointSize(8)
        font.setBold(True)
        font.setWeight(75)
        self.button_add_preset_group_from_selected_rows.setFont(font)
        self.button_add_preset_group_from_selected_rows.setStyleSheet("QPushButton {color:green}")
        self.button_add_preset_group_from_selected_rows.setIconSize(QtCore.QSize(0, 0))
        self.button_add_preset_group_from_selected_rows.setObjectName("button_add_preset_group_from_selected_rows")
        self.horizontalLayout_3.addWidget(self.button_add_preset_group_from_selected_rows)
        self.button_confirm_rows_selection = QtWidgets.QPushButton(self.groups_frame)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.button_confirm_rows_selection.sizePolicy().hasHeightForWidth())
        self.button_confirm_rows_selection.setSizePolicy(sizePolicy)
        self.button_confirm_rows_selection.setMinimumSize(QtCore.QSize(25, 0))
        self.button_confirm_rows_selection.setMaximumSize(QtCore.QSize(170, 16777215))
        font = QtGui.QFont()
        font.setPointSize(8)
        font.setBold(True)
        font.setWeight(75)
        self.button_confirm_rows_selection.setFont(font)
        self.button_confirm_rows_selection.setStyleSheet("QPushButton {color:green}")
        self.button_confirm_rows_selection.setIconSize(QtCore.QSize(0, 0))
        self.button_confirm_rows_selection.setObjectName("button_confirm_rows_selection")
        self.horizontalLayout_3.addWidget(self.button_confirm_rows_selection)
        self.button_cancel_rows_selection = QtWidgets.QPushButton(self.groups_frame)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.button_cancel_rows_selection.sizePolicy().hasHeightForWidth())
        self.button_cancel_rows_selection.setSizePolicy(sizePolicy)
        self.button_cancel_rows_selection.setMinimumSize(QtCore.QSize(25, 0))
        self.button_cancel_rows_selection.setMaximumSize(QtCore.QSize(170, 16777215))
        font = QtGui.QFont()
        font.setPointSize(8)
        font.setBold(True)
        font.setWeight(75)
        self.button_cancel_rows_selection.setFont(font)
        self.button_cancel_rows_selection.setStyleSheet("QPushButton {color:red}")
        self.button_cancel_rows_selection.setIconSize(QtCore.QSize(0, 0))
        self.button_cancel_rows_selection.setObjectName("button_cancel_rows_selection")
        self.horizontalLayout_3.addWidget(self.button_cancel_rows_selection)
        spacerItem6 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_3.addItem(spacerItem6)
        self.gridLayout.addWidget(self.groups_frame, 0, 0, 1, 1)

        self.retranslateUi(Form)
        self.tab_groups.setCurrentIndex(-1)
        self.button_move_top.clicked.connect(Form.move_rows_top)
        self.button_move_up.clicked.connect(Form.move_rows_up)
        self.button_move_down.clicked.connect(Form.move_rows_down)
        self.button_move_bottom.clicked.connect(Form.move_rows_bottom)
        self.button_remove.clicked.connect(Form.remove_rows)
        self.button_add.clicked.connect(Form.add_row)
        self.button_save.clicked.connect(Form.save)
        self.button_load.clicked.connect(Form.load)
        self.button_reset_all.clicked.connect(Form.reset_all)
        self.button_change_color.clicked.connect(Form.select_color)
        self.button_set_value.clicked.connect(Form.set_attribute_value)
        self.button_add_preset_group.clicked.connect(Form.add_preset_group)
        self.button_remove_preset_group.clicked.connect(Form.remove_preset_group)
        self.button_add_preset_group_from_selected_rows.clicked['bool'].connect(Form.add_preset_group_from_selected)
        self.combo_preset.currentIndexChanged['QString'].connect(Form.preset_group_changed)
        self.button_confirm_rows_selection.clicked.connect(Form.confirm_rows_selection)
        self.button_cancel_rows_selection.clicked.connect(Form.cancel_rows_selection)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
        self.button_set_value.setToolTip(_translate("Form", "Set attribute to saved value"))
        self.button_change_color.setToolTip(_translate("Form", "Change row color"))
        self.button_move_top.setToolTip(_translate("Form", "Move row to the top"))
        self.button_move_up.setToolTip(_translate("Form", "Move row up"))
        self.button_remove.setToolTip(_translate("Form", "Remove row"))
        self.button_remove.setText(_translate("Form", "-"))
        self.button_add.setToolTip(_translate("Form", "Add row"))
        self.button_add.setText(_translate("Form", "+"))
        self.button_move_down.setToolTip(_translate("Form", "Move row down"))
        self.button_move_bottom.setToolTip(_translate("Form", "Move row to the bottom"))
        self.button_reset_all.setText(_translate("Form", "Reset"))
        self.button_load.setText(_translate("Form", "Load from file"))
        self.checkbox_allow_double_click.setText(_translate("Form", "Double click sets value"))
        self.checkbox_synced_with_server.setText(_translate("Form", "Synced"))
        self.button_save.setText(_translate("Form", "Save to file"))
        self.label_saved_groups.setText(_translate("Form", "Preset groups:"))
        self.button_remove_preset_group.setToolTip(_translate("Form", "Remove selected group"))
        self.button_remove_preset_group.setText(_translate("Form", "-"))
        self.button_add_preset_group.setToolTip(_translate("Form", "Add new save group"))
        self.button_add_preset_group.setText(_translate("Form", "+ Empty"))
        self.button_add_preset_group_from_selected_rows.setToolTip(_translate("Form", "Add new save group"))
        self.button_add_preset_group_from_selected_rows.setText(_translate("Form", "+ From selected rows"))
        self.button_confirm_rows_selection.setToolTip(_translate("Form", "Add new save group"))
        self.button_confirm_rows_selection.setText(_translate("Form", "Confirm selected rows"))
        self.button_cancel_rows_selection.setToolTip(_translate("Form", "Add new save group"))
        self.button_cancel_rows_selection.setText(_translate("Form", "Cancel"))

from kamzik3 import resource_kamzik3_rc



from kamzik3.constants import *
from kamzik3.devices.device import Device
from kamzik3.devices.observer import Observer


class DeviceChannel(Device, Observer):

    def __init__(self, device, channel, device_id=None, config=None):
        self.configured = False
        self.device = device
        self.channel = channel
        super(DeviceChannel, self).__init__(device_id, config)
        self.set_status(STATUS_CONNECTING)
        self.device.attach_observer(self)

    def command(self, command, callback=None, with_token=False, returning=True):
        return self.device.command(command, callback, with_token, returning)

    def subject_update(self, key, value, subject):
        if key == ATTR_STATUS:
            if value in READY_DEVICE_STATUSES:
                self.handle_configuration()
            else:
                self.stop_polling()
                self.configured = False
                self.set_status(value)

    def poll_command(self, command, interval):
        self.device.poll_command(command, interval)

    def remove_poll_command(self, command, interval):
        self.device.remove_poll_command(command, interval)

    def handle_configuration(self):
        raise NotImplementedError(u"Must be implemented in subclass")

    def disconnect(self):
        self.stop_polling()
        self.configured = False
        self.device.detach_observer(self)
        self.set_status(STATUS_DISCONNECTED)

    def reconnect(self, *args):
        self.device.attach_observer(self)
        self.handle_configuration()



from ctypes import CDLL

from kamzik3 import DeviceError
# from msl.loadlib import LoadLibrary, Server32
from kamzik3.constants import *
from kamzik3.devices.device import Device


class DeviceOsLibrary(Device):

    def __init__(self, os_library, device_id=None, config=None):
        self.os_library = os_library
        self.lib = None
        super().__init__(device_id, config)
        self.connect()

    def _init_attributes(self):
        Device._init_attributes(self)
        self.create_attribute(ATTR_LIBRARY, default_value=self.os_library, readonly=True)

    def connect(self, *args):
        """
        Call only this function to connect devices to port / socket / library / ...
        :param args: connect attributes
        """
        self.connecting = True
        self.handle_configuration_event()
        try:
            self.lib = CDLL(self.os_library)
            # self.lib = LoadLibrary(self.os_library)
        except OSError as e:
            raise DeviceError("Error loading library {}. {}".format(self.os_library, e))
        self.connected = True
        self.connecting = False
        self.set_status(STATUS_IDLE)



import logging

import oyaml
import yaml
from PyQt5.QtCore import pyqtSlot, pyqtSignal
from PyQt5.QtWidgets import QSpacerItem, QSizePolicy, QListWidgetItem

import kamzik3
from kamzik3 import DeviceError
from kamzik3.constants import *
from kamzik3.gui.deviceWidget import DeviceWidget
from kamzik3.gui.macroProgressWidget import MacroProgressWidget
from kamzik3.gui.macroStepWidget import MacroStepWidget
from kamzik3.gui.scanAttributeWidget import ScanAttributeWidget
from kamzik3.gui.templates.macroToolTemplate import Ui_Form
from kamzik3.macro.macro import Macro
from kamzik3.snippets.snippetsGenerators import axis_name_generator
from kamzik3.snippets.snippetsWidgets import clear_layout, show_error_message, show_prompt_dialog, show_info_message, \
    show_question_dialog
from kamzik3.snippets.snippetsYaml import YamlDumper

CUSTOM_TEMPLATE = u"Custom"


class MacroToolWidget(Ui_Form, DeviceWidget):
    sig_reload_templates = pyqtSignal()
    sig_set_macro_prefix = pyqtSignal()
    sig_running_macro_update = pyqtSignal()
    sig_finished_macro_update = pyqtSignal()
    sig_last_error_update = pyqtSignal(str)

    def __init__(self, device, model_image=None, config=None, parent=None):
        self.available_templates = None
        self.config = config
        if self.config is None:
            self.config = {}
        self.iterative_widgets = []
        self.action_widgets = []
        self.displayed_macros = {}
        self.loaded_templates = {}
        self.logger = logging.getLogger("Gui.MacroToolWidget")
        DeviceWidget.__init__(self, device, model_image, config, parent)

    def init_signals(self):
        super().init_signals()
        self.sig_reload_templates.connect(self.slot_reload_templates)
        self.sig_set_macro_prefix.connect(self.slot_set_macro_prefix)
        self.sig_running_macro_update.connect(self.slot_set_running_macros)
        self.sig_finished_macro_update.connect(self.slot_set_finished_macros)
        self.sig_last_error_update.connect(self.slot_handle_last_error)

    @pyqtSlot()
    def slot_handle_configuration(self):
        """
        Handle widget configuration.
        If overloading make sure to set configured flag.
        :return:
        """
        self.device.attach_attribute_callback(ATTR_TEMPLATES_COUNT,
                                              lambda _, __: self.sig_reload_templates.emit())
        self.device.attach_attribute_callback(ATTR_MACRO_PREFIX,
                                              lambda _, __: self.sig_set_macro_prefix.emit())
        self.device.attach_attribute_callback(ATTR_RUNNING_COUNT,
                                              lambda _, __: self.sig_running_macro_update.emit())
        self.device.attach_attribute_callback(ATTR_FINISHED_COUNT,
                                              lambda _, __: self.sig_finished_macro_update.emit())
        self.device.attach_attribute_callback(ATTR_LAST_ERROR,
                                              lambda key, value: self.sig_last_error_update.emit(value))
        self.slot_reload_templates()
        self.slot_set_macro_prefix()
        self.slot_set_running_macros()
        DeviceWidget.slot_handle_configuration(self)

    @pyqtSlot(str)
    def slot_handle_last_error(self, value):
        if len(value) > 0:
            show_error_message(value, self)

    def slot_set_status(self, value):
        if value == STATUS_DISCONNECTED:
            while self.tab_widget_macro_list.count() != 0:
                self.close_macro_tab(0)
        return super().slot_set_status(value)

    @pyqtSlot()
    def slot_reload_templates(self):
        self.loaded_templates = self.device.get_templates()
        self.combobox_template.clear()
        self.combobox_template.blockSignals(True)

        self.combobox_template.addItems(self.loaded_templates.keys())
        self.combobox_template.addItem(CUSTOM_TEMPLATE)

        self.combobox_template.blockSignals(False)
        self.template_selected(self.combobox_template.currentText())

    @pyqtSlot()
    def slot_set_macro_prefix(self):
        macro_preifx = self.device.get_value(ATTR_MACRO_PREFIX)
        self.label_new_macro.setText(u"Create new {}".format(macro_preifx))

    def slot_set_running_macros(self):
        for running_macro in self.device.get_running_macros():
            if running_macro not in self.displayed_macros:
                macro = kamzik3.session.get_device(running_macro)
                template, _ = self.device.get_macro_template(macro_id=running_macro)
                running_macro_widget = MacroProgressWidget(macro, template)
                self.displayed_macros[running_macro] = running_macro_widget
                self.tab_widget_macro_list.addTab(running_macro_widget, running_macro)
                self.tab_widget_macro_list.setCurrentIndex(self.tab_widget_macro_list.count() - 1)

    @pyqtSlot()
    def slot_set_finished_macros(self):
        macros_on_server = self.device.get_finished_macros() + self.device.get_running_macros()
        remove_ids = []
        for widget_index in reversed(range(self.tab_widget_macro_list.count())):
            widget = self.tab_widget_macro_list.widget(widget_index)
            if widget.device.device_id not in macros_on_server:
                remove_ids.append(widget_index)
                self.close_macro_tab(widget_index, False)

    def get_template_meta(self):
        """
        Go thru all widgets in template and export it's meta.
        :return: list of meta data describing template
        """
        meta = []
        layout = self.widget_custom_template.layout()
        for item in range(len(layout)):
            widget = layout.itemAt(item).widget()
            if widget is not None:
                meta.append(widget.get_template())
        return meta

    def export_template(self):
        """
        Export and save new created template.
        :return:
        """
        try:
            items = self.get_template_meta()
        except ValueError as e:
            self.logger.info(e)
            show_error_message(e, self)
            return

        template_title = show_prompt_dialog("Template title")
        if not template_title or template_title == "":
            return

        selected_measurement_groups = [item.text() for item in self.list_measurement_groups.selectedItems()]
        data = {
            "title": template_title,
            "iterations": len(self.iterative_widgets),
            "actions": len(self.action_widgets),
            "items": items,
            "measurement_groups": selected_measurement_groups
        }
        yaml_serialized_data = oyaml.dump(data)
        try:
            template_file = self.device.create_template(template_metadata=yaml_serialized_data)
            show_info_message("New template {} saved.".format(template_file), parent=self)
        except DeviceError as e:
            show_error_message(e, parent=self)

    def remove_template(self):
        """
        Remove currently selected template.
        :return:
        """
        current_template = self.combobox_template.currentText()
        if show_question_dialog("Do You want to remove template {} ?".format(current_template), u"Remove template",
                                self):
            try:
                self.device.remove_template(template_id=current_template)
            except DeviceError as error_message:
                error_message = "Template could not be removed: {}".format(error_message)
                self.logger.error(error_message)
                show_error_message(error_message, self)

    def template_selected(self, template_id):
        if template_id == CUSTOM_TEMPLATE:
            self.reset_template_input()
            self.init_measurement_groups_list()
            self.widget_macro_components.setCurrentWidget(self.page_template_designer)
        else:
            self.reset_macro_input()
            self.widget_macro_components.setCurrentWidget(self.page_macro_components)
            self.generate_macro_layout(self.loaded_templates.get(template_id))

    def reset_template_input(self):
        clear_layout(self.scroll_area_custom_template, self.widget_custom_template.layout())
        self.iterative_widgets = []
        self.action_widgets = []
        self.vertical_spacer = QSpacerItem(40, 20, QSizePolicy.Minimum, QSizePolicy.Expanding)

    def reset_macro_input(self):
        clear_layout(self.scroll_area_macro, self.widget_macro_list.layout())
        self.iterative_widgets = []
        self.action_widgets = []
        self.vertical_spacer = QSpacerItem(40, 20, QSizePolicy.Minimum, QSizePolicy.Expanding)

    def init_measurement_groups_list(self):
        measurement_groups = self.device.get_measurement_groups()
        self.list_measurement_groups.clear()
        for title in measurement_groups.keys():
            self.list_measurement_groups.addItem(QListWidgetItem(title))
        self.list_measurement_groups.setCurrentRow(0)

    def add_action_step(self):
        """
        Add action step widget.
        Which is either set attribute or method.
        :return: None
        """
        self.widget_custom_template.layout().removeItem(self.vertical_spacer)
        step_widget = MacroStepWidget(config={CFG_TEMPLATE: True})
        self.action_widgets.append(step_widget)
        self.widget_custom_template.layout().addWidget(step_widget)
        self.widget_custom_template.layout().addItem(self.vertical_spacer)
        self.update_template_view_step_titles()

    def add_iterative_step(self):
        """
        Add iterative step widget.
        This widget represents any scan over numerical value.
        :return: None
        """
        self.widget_custom_template.layout().removeItem(self.vertical_spacer)
        iterative_widget = ScanAttributeWidget(config={CFG_TEMPLATE: True})
        self.iterative_widgets.append(iterative_widget)
        self.widget_custom_template.layout().addWidget(iterative_widget)
        self.widget_custom_template.layout().addItem(self.vertical_spacer)
        self.update_template_view_step_titles()

    def update_template_view_step_titles(self):
        """
        Since we are using different names for scan widgets like slow, fast scan...
        After adding
        :return:
        """
        axis_names = axis_name_generator(len(self.iterative_widgets))
        for widget in self.iterative_widgets:
            widget.set_title(next(axis_names))

    def generate_macro_layout(self, template):
        if template is None:
            return
        meta, file_path = template
        axis_names = axis_name_generator(meta["iterations"])

        for item in meta.get("items", []):
            if item["type"] == "iterative":
                iterative_widget = ScanAttributeWidget("{} {}".format(next(axis_names), item["attribute"]), config=item)
                self.iterative_widgets.append(iterative_widget)
                self.widget_macro_list.layout().addWidget(iterative_widget)
            elif item["type"] == "action":
                step_title = None
                if item["step"] == "Set attribute":
                    step_title = "{} {}".format(item["step"], item["attribute"])
                elif item["step"] == "Execute method":
                    step_title = "{} {}".format(item["step"], item["method"])
                action_widget = MacroStepWidget(step_title, config=item)
                self.action_widgets.append(action_widget)
                self.widget_macro_list.layout().addWidget(action_widget)

        self.widget_macro_list.layout().addItem(self.vertical_spacer)

    def save(self):
        current_template, template_filepath = self.loaded_templates.get(self.combobox_template.currentText())
        scan_layout = self.widget_macro_list.layout()
        # Create macro for current scan
        macro = Macro("Macro", repeat_count=self.input_repeat_macro.value(), comment=self.input_comment.text())
        try:
            # Got thru current template and create all steps
            for index, item in enumerate(current_template["items"]):
                scan_item_widget = scan_layout.itemAt(index).widget()
                step = scan_item_widget.save()
                macro.add(step)
        except (ValueError, KeyError) as e:
            show_error_message(e, self)
            macro.remove()
            return None

        return macro

    def start_macro(self):
        template_id = self.combobox_template.currentText()
        macro = self.save()
        if macro is None:
            return
        yaml_serialized_macro = yaml.dump(macro, Dumper=YamlDumper)
        try:
            macro_id = self.device.upload_macro(template_id=template_id, macro_data=yaml_serialized_macro)
            self.device.start_macro(macro_id=macro_id)
        except DeviceError as e:
            show_error_message(e, self)

    @pyqtSlot(int)
    def close_macro_tab(self, tab_index, remove_tab=True):
        macro_widget = self.tab_widget_macro_list.widget(tab_index)
        macro_name = macro_widget.device.device_id
        del self.displayed_macros[macro_name]
        if remove_tab:
            self.tab_widget_macro_list.removeTab(tab_index)
        macro_widget.close()



import logging
import time
from logging.handlers import TimedRotatingFileHandler

import numpy as np

from kamzik3 import DeviceError, DeviceUnknownError, units
from kamzik3.constants import *
from kamzik3.devices.attribute import Attribute
from kamzik3.devices.device import Device
from kamzik3.snippets.snippetLogging import set_rotating_file_handler, set_file_handler
from kamzik3.snippets.snippetsTimer import PreciseCallbackTimer

ATTR_LOGFILE = "Logfile"
ATTR_HEADER = "Header"
ATTR_LAST_LOG_LINE = "Last log line"
ATTR_INTERVAL = "Interval"
ATTR_LOGGING = "Logging"


class AttributeLogger(Device):
    # Logging interval in ms
    interval = 1e3
    separator = ";"
    logged_attributes = None
    attribute_logger = None
    attribute_logger_handler = None
    preset_header = u""
    logger_ticker = None

    def __init__(self, log_file_name, device_id=None, config=None):
        if self.logged_attributes is None:
            self.logged_attributes = []

        self.log_file_name = log_file_name
        self.log_formatter = logging.Formatter(self.separator.join(['%(asctime)s', '%(created)s', '%(message)s']),
                                               datefmt='%Y-%m-%d %H:%M:%S')
        Device.__init__(self, device_id, config)
        self.connect()

    def connect(self):
        self.handle_connect_event()
        self.set_status(STATUS_IDLE)

    def _init_attributes(self):
        super(AttributeLogger, self)._init_attributes()
        self.create_attribute(ATTR_HEADER, readonly=True)
        self.create_attribute(ATTR_INTERVAL, default_value=self.interval, min_value=100, max_value=3600 * 1e3,
                              unit=u"ms", default_type=np.uint32, set_function=self._set_timeout)
        self.create_attribute(ATTR_LAST_LOG_LINE, readonly=True)
        self.create_attribute(ATTR_LOGFILE, default_value=self.log_file_name, readonly=True)
        self.create_attribute(ATTR_LOGGING, min_value=0, max_value=1, default_type=np.bool, set_function=self.start)

    def command(self, command, callback=None, with_token=False, returning=True):
        raise DeviceError("Device does not accept any commands.")

    def _set_timeout(self, timeout):
        self.timeout = timeout * 1e-3

    def _get_logged_attributes(self):
        logged_values = []
        value_unit = u""
        for device_id, attribute in self.logged_attributes[:]:
            try:
                device = self.session.get_device(device_id)
                logged_value, value_unit = None, u""
                if device.in_statuses(READY_DEVICE_STATUSES):
                    attribute = device.get_attribute(attribute)
                    if attribute is not None and attribute.value() is not None:
                        attribute_value = attribute.value()
                        if isinstance(attribute_value, units.Quantity):
                            logged_value = "{:~}".format(attribute.value())
                        else:
                            logged_value = attribute.value()
            except DeviceUnknownError:
                logged_value = None

            if logged_value is None:
                logged_value, value_unit = u"None", u""

            logged_values.append("{}{}".format(logged_value, value_unit))

        return logged_values

    def set_log_file_name(self, log_file_name):
        self.set_raw_value(ATTR_LOGFILE, log_file_name)
        attribute_logger = logging.getLogger(u"AttributeLogger.{}".format(self.device_id))
        if self.attribute_logger_handler is not None:
            attribute_logger.removeHandler(self.attribute_logger_handler)

        if self.config.get("rotating", True):
            self.attribute_logger_handler = set_rotating_file_handler(attribute_logger, self.log_file_name,
                                                                      self.log_formatter)

            def do_rollover():
                TimedRotatingFileHandler.doRollover(self.attribute_logger_handler)
                self.write_header()

            self.attribute_logger_handler.doRollover = do_rollover
        else:
            self.attribute_logger_handler = set_file_handler(attribute_logger, self.log_file_name,
                                                             self.log_formatter)
        attribute_logger.setLevel(logging.INFO)
        return attribute_logger

    def handle_configuration(self):
        start_at = time.time()

        def _finish_configuration(*_, **__):
            self._config_commands()
            self._config_attributes()
            self.attribute_logger = self.set_log_file_name(self.log_file_name)
            self.set_status(STATUS_CONFIGURED)
            self.logger.info(u"Device configuration took {} sec.".format(time.time() - start_at))

        _finish_configuration()

    def handle_interval_timeout(self):
        logged_values = self._get_logged_attributes()
        last_log_line = self.separator.join(logged_values)
        self.set_raw_value(ATTR_LAST_LOG_LINE, last_log_line)
        if self.attribute_logger:
            self.attribute_logger.info(last_log_line)

    def add_logged_attribute(self, device_id, attribute):
        attribute = Attribute.list_attribute(attribute)
        item = (device_id, attribute)

        if item not in self.logged_attributes:
            self.logged_attributes.append(item)

    def remove_logged_attribute(self, device_id, attribute):
        if isinstance(attribute, str):
            attribute = [attribute]
        elif isinstance(attribute, tuple):
            attribute = list(attribute)

        self.logged_attributes.remove((device_id, attribute))

    def generate_header(self):
        header_line = ["Datetime", "Timestamp"]
        for device_id, attribute in self.logged_attributes[:]:
            header_line.append(u"{} {}".format(device_id, attribute))
        return "# " + self.separator.join(header_line).replace("'", "").replace(", ", ",")

    def write_header(self):
        header = self.preset_header + self.generate_header()
        self.set_raw_value(ATTR_HEADER, header)
        with open(self.log_file_name, "a+") as fp:
            fp.write(header)
            fp.write("\n")

    def start(self, flag=True):
        if flag:
            if self.logger_ticker is not None and self.is_status(STATUS_BUSY):
                self.logger_ticker.stop()
                raise DeviceError(u"Logging is already running")

            self.logger_ticker = PreciseCallbackTimer(self.interval, self.handle_interval_timeout, with_correction=True)
            self.write_header()
            self.set_status(STATUS_BUSY)
            self.logger_ticker.start()
        else:
            self.stop()

    def stop(self):
        try:
            self.logger_ticker.stop()
            self.set_status(STATUS_IDLE)
        except TypeError:
            pass

    def close(self):
        self.stop()
        super(AttributeLogger, self).close()

    def disconnect(self):
        # Remove all handlers
        self.stop()
        try:
            for handler in self.attribute_logger.handlers[:]:
                handler.close()
        except AttributeError:
            # No attribute_logger in which case just continue
            pass
        self.logged_attributes = None
        self.attribute_logger = None
        self.attribute_logger_handler = None
        self.logged_attributes = []
        self.log_formatter = None
        return super().disconnect()


class AttributeLoggerTriggered(AttributeLogger):

    def start(self, flag=True):
        if flag:
            if self.is_status(STATUS_BUSY):
                raise DeviceError(u"Logging is already running")

            self.write_header()
            self.set_status(STATUS_BUSY)
        else:
            self.stop()

    def trigger(self):
        self.handle_interval_timeout()

    def stop(self):
        try:
            self.set_status(STATUS_IDLE)
        except TypeError:
            pass



from PyQt5.QtCore import QObject, QPointF, QPoint, Qt
from PyQt5.QtGui import QBrush

NORMAL_TEXT = 0
STATIC_TEXT = 1


class GlWidget(QObject):
    all = {}
    tooltipText = None
    text = None
    textCoords = None
    opacity = 1
    onClick = None
    angle = None
    background = None

    def __init__(self, parent, widget_id, coords, texture=None, border=False):
        QObject.__init__(self)
        self.setParent(parent)
        self.font = None
        self.add_border = border
        self.id = widget_id
        self.coords = coords
        self.texture = texture
        self.clickPath = coords
        self.clickPathTransformed = coords
        self.align = Qt.AlignCenter
        self.background_shape = "rect"
        self.background_coords = coords
        GlWidget.all[self.id] = self

    def set_background(self, coords=None, color=Qt.white, shape=None):
        if coords is None:
            coords = self.background_coords
        else:
            self.background_coords = coords
        if shape is None:
            shape = self.background_shape
        else:
            self.background_shape = shape
        self.background = (coords, color, shape)
        return self

    def set_text(self, text, font, align=Qt.AlignCenter):
        self.text = text
        self.font = font
        self.align = align
        return self

    def set_tooltip_text(self, value):
        self.tooltipText = value
        return self

    def draw(self, painter):
        painter.save()
        if self.angle is not None:
            painter.translate(self.coords.center())
            painter.rotate(self.angle)
            painter.translate(self.coords.center() * -1)

        if self.texture is not None:
            painter.setOpacity(self.opacity)
            painter.drawPixmap(self.coords, self.texture)
        if self.background is not None:
            coords, color, shape = self.background
            pen = painter.pen()
            brush = painter.brush()
            if not self.add_border:
                painter.setPen(Qt.NoPen)
            painter.setBrush(QBrush(color))
            if shape == "rect":
                painter.drawRect(coords)
            elif shape == "ellipse":
                painter.drawEllipse(coords)
            painter.setPen(pen)
            painter.setBrush(brush)

        self.matrix = painter.worldTransform()
        self.clickPathTransformed = self.matrix.mapToPolygon(self.clickPath)

        painter.setOpacity(1)
        self.paint_text(painter)
        painter.restore()
        return self

    def paint_text(self, painter):
        if self.text is None:
            return
        painter.setFont(self.font)
        painter.drawText(self.coords, self.align, self.text)
        return self

    @staticmethod
    def _getStaticTextCoordds(boundRect, staticText):
        center = boundRect.center()
        font_size = staticText.size()
        return QPointF(center.x() - (font_size.width() / 2.), center.y() - font_size.height() / 2.)

    @staticmethod
    def get(widget_id):
        return GlWidget.all.get(widget_id)

    def set_on_click(self, callback):
        self.onClick = callback
        return self

    def click(self):
        self.onClick(self)

    def rotate(self, angle):
        self.angle = angle
        return self

    def set_click_path(self, click_path):
        self.clickPath = click_path
        return self

    def is_point_inside(self, point):
        if isinstance(point, (QPoint, QPointF)):
            return self.clickPathTransformed.containsPoint(point, Qt.FillRule(Qt.OddEvenFill))


class GlSwitchTexture(GlWidget):
    state = 0

    def __init__(self, parent, widget_id, coords, on_texture, off_texture):
        GlWidget.__init__(self, parent, widget_id, coords)
        self.on_texture = on_texture
        self.off_texture = off_texture

    def set_state(self, state):
        self.state = state
        return self

    def draw(self, painter):
        if self.state:
            self.texture = self.on_texture
        else:
            self.texture = self.off_texture
        return GlWidget.draw(self, painter)


class GlSwitchColor(GlWidget):
    state = 0

    def __init__(self, parent, widget_id, coords, on_color, off_color):
        GlWidget.__init__(self, parent, widget_id, coords)
        self.on_color = on_color
        self.off_color = off_color

    def set_state(self, state):
        self.state = state
        return self

    def draw(self, painter):
        if self.background is not None:
            coords, _, shape = self.background
        else:
            coords, shape = self.coords, "rect"
        if self.state:
            self.background = (coords, self.on_color, shape)
        else:
            self.background = (coords, self.off_color, shape)
        return GlWidget.draw(self, painter)



import re
import time

import numpy as np
import serial
from bidict import frozenbidict
from pint.errors import UndefinedUnitError

from kamzik3 import DeviceUnitError
from kamzik3.constants import *
from kamzik3.devices.deviceChannel import DeviceChannel
from kamzik3.devices.devicePort import DevicePort
from kamzik3.devices.deviceSocket import DeviceSocket
from kamzik3.snippets.snippetsDecorators import expose_method
from kamzik3.snippets.snippetsUnits import device_units

""" Example of yaml configuration
Smaract0: &Smaract0 !Device:kamzik3.devices.deviceSmaractMcs1.DeviceSmaractMcs1
    device_id: Smaract0
    host: 192.168.83.91
Smaract0_axis_0: &Smaract0_axis_0 !Device:kamzik3.devices.deviceSmaractMcs1.DeviceSmaractMcs1Channel
    device: *Smaract0
    channel: 0
    device_id: MotorAxis0
    config:
      attributes:
        !!python/tuple [Sensor type, Value]: 1
        !!python/tuple [Position +limit, Value]: 15mm
        !!python/tuple [Position -limit, Value]: -15mm
        !!python/tuple [Position, Tolerance]: [10nm, 10nm]
"""


class DeviceSmaractMcs1(DeviceSocket):
    push_buffer_size = 50
    response_timeout = 3000
    terminator = b"\n"
    handtool_modes = frozenbidict({"Disabled": 0, "Enabled": 1, "Display only": 2})
    sensor_modes = frozenbidict({"Disabled": 0, "Enabled": 1, "Power safe": 2})
    channel_command_pattern = re.compile(":([A-Z]+)([-0-9]+),(.*)")
    linear_sensor_type = [None, 1, 5, 6, 9, 18, 21, 24, 32, 38, 40, 41, 42, 43, 44]
    sensor_types = [1, 2, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 23, 24, 32, 37, 38, 39, 40,
                    41, 42, 43, 44, 48, 49, 50, 51, 53, 54]
    safe_direction_modes = frozenbidict({"Forward": 0, "Backward": 1})

    def __init__(self, host, port=5000, device_id=None, config=None):
        super(DeviceSmaractMcs1, self).__init__(host, port, device_id, config)

    def _init_attributes(self):
        super(DeviceSmaractMcs1, self)._init_attributes()
        self.create_attribute(ATTR_STATE, readonly=True, default_type=np.uint16)
        self.create_attribute(ATTR_SERIAL_NUMBER, readonly=True)
        self.create_attribute(ATTR_CHANNELS, readonly=True, default_type=np.uint16)
        self.create_attribute(ATTR_HAND_TOOL_MODE, readonly=True, default_type=np.uint16)
        self.create_attribute(ATTR_HAND_TOOL_MODE, default_value="Display only",
                              default_type=self.handtool_modes.keys(),
                              set_function=self.set_hand_tool_mode)
        self.create_attribute(ATTR_SENSOR_MODE, default_type=self.sensor_modes.keys(),
                              set_function=self.set_sensor_mode)
        self.create_attribute(ATTR_SENSOR_PRESENT, default_value=True, readonly=True, default_type=np.bool)

    def valid_command_format(self, command):
        if command is not None and command[-1] == "\n":
            return True
        else:
            return False

    def handle_configuration(self):
        start_at = time.time()

        def _finish_configuration(*_, **__):
            self._config_commands()
            self._config_attributes()
            self.set_status(STATUS_CONFIGURED)
            self.logger.info(u"Device configuration took {} sec.".format(time.time() - start_at))

        # Set communication to synchronous
        self.command(u":SCM0\n")
        # Set hand-tool mode to just display positions
        self.set_hand_tool_mode("Display only")
        # Get Interface version
        self.command(u":GSI\n")
        # Get number of channels
        self.command(u":GNC\n")
        # Get Interface version
        self.command(u":GSE\n", callback=_finish_configuration)

    def handle_readout(self, readout_buffer):
        command, output, callback, token = super(DeviceSmaractMcs1, self).handle_readout(readout_buffer)

        if output[:2] == ":E":
            channel, error_number = [int(v) for v in output[2:].split(",")]
            if error_number > 0:
                # if error_number == 129:
                #     self.set_raw_value(ATTR_SENSOR_PRESENT, False)
                #     self.set_raw_value(ATTR_SENSOR_MODE, "Disabled")
                # for channel in range(self[ATTR_CHANNELS][VALUE]):
                #     self.notify((channel, ATTR_SENSOR_MODE), "Disabled")
                #     self.notify((channel, ATTR_SENSOR_PRESENT), False)
                # elif error_number == 140:
                #     self.set_raw_value(ATTR_SENSOR_MODE, "Disabled")
                # for channel in range(self[ATTR_CHANNELS][VALUE]):
                #     self.notify((channel, ATTR_SENSOR_MODE), "Disabled")
                self.handle_command_error(command, output)
        elif command == ":GSI\n":
            self.set_raw_value(ATTR_SERIAL_NUMBER, output.strip(":IN"))
        elif command == ":GNC\n":
            channels = int(output.strip(":N"))
            self.set_raw_value(ATTR_CHANNELS, channels)
        elif command == ":GSE\n":
            mode = self.sensor_modes.inv[int(output.strip(":SE"))]
            self.set_raw_value(ATTR_SENSOR_MODE, mode)
            for channel in range(self[ATTR_CHANNELS][VALUE]):
                self.notify((channel, ATTR_SENSOR_MODE), mode)
        elif output[:3] == ":BR":
            pass
        else:
            find = self.channel_command_pattern.search(output)
            if find is not None and find != -1:
                attribute = None
                cmd, channel, value = find.group(1), int(find.group(2)), find.group(3)
                if cmd == "S":
                    attribute, value = ATTR_STATE, int(value)
                elif cmd == "P":
                    attribute, value = ATTR_POSITION, int(value)
                elif cmd == "A":
                    position, revolutions = value.split(",")
                    self.notify((channel, ATTR_REVOLUTIONS), int(revolutions))
                    attribute, value = ATTR_POSITION, int(position)
                elif cmd == "ST":
                    attribute, value = ATTR_SENSOR_TYPE, int(value)
                elif cmd == "PPK":
                    attribute, value = ATTR_REFERENCED, bool(int(value))
                elif cmd == "SD":
                    safe_direction = self.safe_direction_modes.inv[int(value)]
                    attribute, value = ATTR_SAFE_DIRECTION, safe_direction
                elif cmd == "CLA":
                    attribute, value = ATTR_ACCELERATION, int(value)
                elif cmd == "CLS":
                    attribute, value = ATTR_VELOCITY, int(value)
                elif cmd == "AL":
                    min_position, min_revolution, max_position, max_revolution = [int(i) for i in value.split(",")]
                    attribute, value = ATTR_POSITION_LOWER_LIMIT, (min_position, min_revolution)
                    self.notify((channel, ATTR_POSITION_UPPER_LIMIT), (max_position, max_revolution))
                elif cmd == "PL":
                    min_position, max_position = value.split(",")
                    attribute, value = ATTR_POSITION_LOWER_LIMIT, int(min_position)
                    self.notify((channel, ATTR_POSITION_UPPER_LIMIT), int(max_position))
                if attribute:
                    self.notify((channel, attribute), value)

        if callback is not None:
            self.handle_readout_callback(callback, command, readout_buffer)
        if token:
            self.device_server.push_message(self.device_id, [command, readout_buffer], token)
            for observer in self._observers[:]:
                if isinstance(observer, DeviceSmaractMcs1Channel):
                    self.device_server.push_message(observer.device_id, [command, readout_buffer], token)

        # if callback is not None:
        #     self.handle_readout_callback(callback, command, readout_buffer)
        # if token:
        #     self.device_server.push_message(self.device_id, readout_buffer, token)
        #
        #     for observer in self._observers[:]:
        #         if isinstance(observer, DeviceSmaractMcs1Channel):
        #             self.device_server.push_message(observer.device_id, readout_buffer, token)

    def set_hand_tool_mode(self, value):
        self.logger.info(u"Set hand tool mode to {}".format(value))
        int_mode = self.handtool_modes[value]
        self.command(u":SHE{}\n".format(int_mode))

    def set_sensor_mode(self, value):
        self.logger.info(u"Set sensor mode to {}".format(value))
        for channel in range(self[ATTR_CHANNELS][VALUE]):
            self.notify((channel, ATTR_SENSOR_MODE), value)

        value = self.sensor_modes[value]
        self.command(u":SSE{}\n".format(value))


class DeviceSmaractMcs1Port(DevicePort, DeviceSmaractMcs1):
    push_buffer_size = DeviceSmaractMcs1.push_buffer_size
    response_timeout = DeviceSmaractMcs1.response_timeout
    terminator = DeviceSmaractMcs1.terminator

    def __init__(self, port, baud_rate=9600, parity=serial.PARITY_NONE, stop_bits=serial.STOPBITS_ONE,
                 byte_size=serial.EIGHTBITS, device_id=None, config=None):
        super(DeviceSmaractMcs1Port, self).__init__(port, baud_rate, parity, stop_bits, byte_size, device_id, config)

    def _init_attributes(self):
        super(DeviceSmaractMcs1Port, self)._init_attributes()
        self.create_attribute(ATTR_BAUDRATE, default_value=self.baud_rate, readonly=False, default_type=self.BAUDRATES,
                              set_function=self.set_baudrate)

    def handshake(self):
        self.logger.info(u"Handshake initiated")
        self.push(b":SCM0\n")
        time.sleep(0.05)  # Give some time for devices to reply
        if self.serial_port.in_waiting != 0:
            if self.read_all().strip() == ":E-1,0":
                return True
        return False

    def set_baudrate(self, value):
        self.logger.info(u"Set baud-rate to {}".format(value))
        self.command(u":CB{}\n".format(int(value)))


class DeviceSmaractMcs1Channel(DeviceChannel):
    sensor_types = DeviceSmaractMcs1.sensor_types
    linear_sensor_type = DeviceSmaractMcs1.linear_sensor_type
    sensor_modes = DeviceSmaractMcs1.sensor_modes
    safe_direction_modes = DeviceSmaractMcs1.safe_direction_modes

    def _init_attributes(self):
        super(DeviceSmaractMcs1Channel, self)._init_attributes()
        self.create_attribute(ATTR_STATE, readonly=True, default_type=np.uint16)
        self.create_attribute(ATTR_MOVING, readonly=True, default_type=np.bool)
        self.create_attribute(ATTR_POSITION, default_type=np.int64, set_function=self._move_absolute,
                              set_value_when_set_function=False)
        self.create_attribute(ATTR_REVOLUTIONS, readonly=True, default_type=np.int64)
        self.create_attribute(ATTR_CALIBRATING, readonly=True, default_type=np.bool)
        self.create_attribute(ATTR_REFERENCING, readonly=True, default_type=np.bool)
        self.create_attribute(ATTR_CALIBRATED, readonly=True, default_type=np.bool)
        self.create_attribute(ATTR_REFERENCED, readonly=True, default_type=np.bool)
        self.create_attribute(ATTR_MOVEMENT_LOCKED, readonly=True, default_type=np.bool)
        self.create_attribute(ATTR_SENSOR_MODE, readonly=True, default_type=self.sensor_modes.keys())
        self.create_attribute(ATTR_SENSOR_TYPE, default_type=self.sensor_types, set_function=self.set_sensor_type)
        self.create_attribute(ATTR_SENSOR_PRESENT, readonly=True, default_type=np.bool)
        self.create_attribute(ATTR_ENCODER_BASE_UNIT, readonly=True)
        self.create_attribute(ATTR_ENCODER_RESOLUTION, readonly=True, default_type=np.int8)
        self.create_attribute(ATTR_POSITION_LOWER_LIMIT, default_type=np.int64,
                              set_function=self.set_lower_position_limit)
        self.create_attribute(ATTR_POSITION_UPPER_LIMIT, default_type=np.int64,
                              set_function=self.set_upper_position_limit)
        self.create_attribute(ATTR_STEP_FREQUENCY, default_value=1000, default_type=np.uint32, min_value=1,
                              max_value=18.5e3, unit="Hz",
                              set_function=self.set_step_frequency)
        self.create_attribute(ATTR_STEP_AMPLITUDE, default_value=1000, default_type=np.uint32, min_value=1,
                              max_value=65535,
                              set_function=self.set_step_amplitude)
        self.create_attribute(ATTR_ACCELERATION, default_type=np.int64, min_value=0, max_value=100e9,
                              set_function=self.set_acceleration)
        self.create_attribute(ATTR_VELOCITY, default_type=np.int64, min_value=0, max_value=1e8,
                              set_function=self.set_velocity)
        self.create_attribute(ATTR_SAFE_DIRECTION, default_type=self.safe_direction_modes.keys(),
                              set_function=self.set_safe_direction)
        self.create_attribute(ATTR_HOLDING_TIME, default_value=0, default_type=np.uint16, min_value=0, max_value=60e3,
                              unit=u"ms")

    def handle_configuration(self):
        if self.configured:
            return

        start_at = time.time()

        def _finish_configuration(*_, **__):
            self._config_commands()
            self._config_attributes()
            self.get_position_limit_range()
            self.start_polling()
            self.set_status(STATUS_CONFIGURED)
            self.logger.info(u"Device configuration took {} sec.".format(time.time() - start_at))

        self.connected = True
        self.get_sensor_type()
        self.get_acceleration()
        self.get_safe_direction()
        self.command(u":GCLS{}\n".format(self.channel), callback=_finish_configuration)

    def subject_update(self, key, value, subject):
        super(DeviceSmaractMcs1Channel, self).subject_update(key, value, subject)

        if self.connected and isinstance(key, tuple) and key[0] == self.channel:
            attribute = key[1]
            if attribute == ATTR_STATE and self.get_value(ATTR_STATE) != value:
                if value in (0, 3):
                    self.set_status(STATUS_IDLE)
                    self.set_value(ATTR_CALIBRATING, False)
                    self.set_value(ATTR_REFERENCING, False)
                    self.set_value(ATTR_MOVING, False)
                else:
                    self.set_status(STATUS_BUSY)
                    if value in (1, 2, 4, 6, 7):
                        self.set_value(ATTR_MOVING, True)

                        if value == 6:
                            self.set_value(ATTR_CALIBRATING, True)
                        elif value == 7:
                            self.set_value(ATTR_REFERENCING, True)
                        elif value == 9:
                            self.set_value(ATTR_MOVEMENT_LOCKED, True)
            elif attribute == ATTR_SENSOR_TYPE:
                self._set_position_unit(value)
                self._set_position_polling(value)
                self.get_referenced()
                self.get_position_limit_range(value)
            elif attribute == ATTR_SENSOR_PRESENT:
                if not value:
                    self.remove_poll_command(":GP{}\n".format(self.channel), 200)
                    self.remove_poll_command(":GA{}\n".format(self.channel), 200)
                else:
                    self._set_position_polling(self[ATTR_SENSOR_TYPE][VALUE])
            elif attribute == ATTR_SENSOR_MODE:
                if value == "Disabled":
                    self.remove_poll_command(":GP{}\n".format(self.channel), 200)
                    self.remove_poll_command(":GA{}\n".format(self.channel), 200)
                else:
                    self._set_position_polling(self[ATTR_SENSOR_TYPE][VALUE])
            elif attribute == ATTR_POSITION and self[ATTR_REVOLUTIONS][VALUE] != 0:
                revolution = 360 * 10 ** abs(self[ATTR_ENCODER_RESOLUTION][VALUE])
                value += self[ATTR_REVOLUTIONS][VALUE] * revolution
            elif attribute == ATTR_POSITION_LOWER_LIMIT and self[ATTR_SENSOR_TYPE][
                VALUE] not in self.linear_sensor_type:
                revolution = 360 * 10 ** abs(self[ATTR_ENCODER_RESOLUTION][VALUE])
                min_position, min_revolutions = value
                value = min_position + min_revolutions * revolution
                # self.set_attribute([ATTR_POSITION, MIN], value)
            elif attribute == ATTR_POSITION_UPPER_LIMIT and self[ATTR_SENSOR_TYPE][
                VALUE] not in self.linear_sensor_type:
                revolution = 360 * 10 ** abs(self[ATTR_ENCODER_RESOLUTION][VALUE])
                max_position, max_revolutions = value
                value = max_position + max_revolutions * revolution
                # self.set_attribute([ATTR_POSITION, MAX], value)
            self.set_value(ATTR_LATENCY, self.device.get_value(ATTR_LATENCY))
            self.set_raw_value(attribute, value)

    def _set_position_unit(self, sensor_type):
        if sensor_type is None:
            return

        prefix = {-12: "p", -9: "n", -6: "u", -3: "m"}
        if sensor_type in self.linear_sensor_type:
            resolution = -9
            unit = "m"
        else:
            resolution = -6
            unit = "deg"
        base_position_unit = prefix.get(resolution, "") + unit
        self.set_attribute((ATTR_ENCODER_RESOLUTION, VALUE), resolution)
        self.set_attribute((ATTR_ENCODER_BASE_UNIT, VALUE), unit)
        self.set_attribute((ATTR_POSITION, UNIT), base_position_unit)
        self.set_attribute((ATTR_POSITION_UPPER_LIMIT, UNIT), base_position_unit)
        self.set_attribute((ATTR_POSITION_LOWER_LIMIT, UNIT), base_position_unit)
        self.set_attribute((ATTR_ACCELERATION, UNIT), "{}/s".format(base_position_unit))
        self.set_attribute((ATTR_VELOCITY, UNIT), "{}/s".format(base_position_unit))

    def _set_position_polling(self, sensor_type):
        if sensor_type is None:
            return
        elif sensor_type in self.linear_sensor_type:
            self.poll_command(":GP{}\n".format(self.channel), 200)
            self.set_value(ATTR_REVOLUTIONS, 0)
        else:
            self.poll_command(":GA{}\n".format(self.channel), 200)

    def start_polling(self):
        super(DeviceSmaractMcs1Channel, self).start_polling()
        self.poll_command(":GS{}\n".format(self.channel), 200)

    def stop_polling(self):
        self.remove_poll_command(":GS{}\n".format(self.channel), 200)
        self.remove_poll_command(":GP{}\n".format(self.channel), 200)
        self.remove_poll_command(":GA{}\n".format(self.channel), 200)

    def _move_absolute(self, value, callback=None):
        self.logger.info(u"Move to position {}".format(value))
        holding_time = self.get_value(ATTR_HOLDING_TIME)
        if self[ATTR_SENSOR_TYPE][VALUE] in self.linear_sensor_type:
            return self.command(u":MPA{},{},{}\n".format(self.channel, int(value), holding_time), callback,
                                with_token=True)
        else:
            revolution = 360 * 10 ** abs(self[ATTR_ENCODER_RESOLUTION][VALUE])
            revolutions = int(value // revolution)
            angular_position = int(value - revolutions * revolution)
            return self.command(u":MAA{},{},{},{}\n".format(self.channel, angular_position, revolutions, holding_time),
                                callback,
                                with_token=True)

    def _move_relative(self, value, callback=None):
        self.logger.info(u"Move by relative position {}".format(value))
        holding_time = self.get_value(ATTR_HOLDING_TIME)
        if self[ATTR_SENSOR_TYPE][VALUE] in self.linear_sensor_type:
            return self.command(u":MPR{},{},{}\n".format(self.channel, int(value), holding_time), callback,
                                with_token=True)
        else:
            revolution = 360 * 10 ** abs(self[ATTR_ENCODER_RESOLUTION][VALUE])
            revolutions = int(value // revolution)
            angular_position = int(value - revolutions * revolution)
            return self.command(u":MAR{},{},{},{}\n".format(self.channel, angular_position, revolutions, holding_time),
                                callback,
                                with_token=True)

    @expose_method()
    def stop(self, callback=None):
        self.logger.info(u"Stop movement")
        return self.command(u":S{}\n".format(self.channel), callback, with_token=True)

    @expose_method()
    def calibrate(self, callback=None):
        self.logger.info(u"Calibrate")
        return self.command(u":CS{}\n".format(self.channel), callback, with_token=True)

    @expose_method()
    def find_reference(self, callback=None):
        self.logger.info(u"Reference")
        reference_direction = self.safe_direction_modes[self.get_value(ATTR_SAFE_DIRECTION)]
        return self.command(u":FRM{},{},0,1\n".format(self.channel, reference_direction), callback, with_token=True)

    @expose_method({"position": ATTR_POSITION})
    def move_absolute(self, position):
        try:
            position = device_units(self, ATTR_POSITION, position)
            self.logger.info(u"Move to absolute position {}".format(position))
            return self.set_value(ATTR_POSITION, position)
        except (UndefinedUnitError, ValueError) as e:
            self.logger.exception(e)
            raise DeviceUnitError(e)

    @expose_method({"position": ATTR_POSITION})
    def move_relative(self, position):
        try:
            position = device_units(self, ATTR_POSITION, position)
            self.logger.info(u"Move by relative position {}".format(position))
            return self.set_value(ATTR_POSITION, self[ATTR_POSITION].value() + position)
        except (UndefinedUnitError, ValueError) as e:
            self.logger.exception(e)
            raise DeviceUnitError(e)

    @expose_method({"steps": "Number of steps", "amplitude": ATTR_STEP_AMPLITUDE, "frequency": ATTR_STEP_FREQUENCY})
    def move_step(self, steps, amplitude=None, frequency=None, callback=None):
        if amplitude in (None, ""):
            amplitude = self[ATTR_STEP_AMPLITUDE][VALUE]
        else:
            amplitude = device_units(self, ATTR_STEP_AMPLITUDE, amplitude).m
        if frequency in (None, ""):
            frequency = self[ATTR_STEP_FREQUENCY][VALUE]
        else:
            frequency = device_units(self, ATTR_STEP_FREQUENCY, frequency).m
        amplitude = int(self[ATTR_STEP_AMPLITUDE].remove_offset_factor(amplitude))
        frequency = int(self[ATTR_STEP_FREQUENCY].remove_offset_factor(frequency))
        self.logger.info(u"Move by {} steps with amplitude {} and frequency of {} Hz".format(int(steps), int(amplitude),
                                                                                             int(frequency)))
        return self.command(u":MST{},{},{},{}\n".format(self.channel, int(steps), amplitude, frequency), callback,
                            with_token=True)

    def get_sensor_type(self):
        self.command(u":GST{}\n".format(self.channel))

    def get_position_limit_range(self, sensor_type=None):
        if sensor_type in self.linear_sensor_type:
            self.command(u":GPL{}\n".format(self.channel))
        else:
            self.command(u":GAL{}\n".format(self.channel))

    def get_velocity(self):
        self.command(u":GCLS{}\n".format(self.channel))

    def get_acceleration(self):
        self.command(u":GCLA{}\n".format(self.channel))

    def get_safe_direction(self):
        self.command(u":GSD{}\n".format(self.channel))

    def get_referenced(self):
        self.command(u":GPPK{}\n".format(self.channel))

    def set_sensor_type(self, value, callback=None):
        self.logger.info(u"Set sensor type to {}".format(value))
        if value != self.get_value(ATTR_SENSOR_TYPE):
            self.remove_poll_command(":GP{}\n".format(self.channel), 200)
            self.remove_poll_command(":GA{}\n".format(self.channel), 200)

            token = self.command(u":SST{},{}\n".format(self.channel, int(value)), callback, with_token=True)
            self.get_sensor_type()
            return token

    def set_lower_position_limit(self, value, callback=None):
        self.logger.info(u"Set lower position limit to {}".format(value))
        upper_limit = self[ATTR_POSITION_UPPER_LIMIT][VALUE]
        # self.set_value(ATTR_POSITION_LOWER_LIMIT, int(value))
        if upper_limit is None:
            upper_limit = value
            # self.set_value(ATTR_POSITION_UPPER_LIMIT, upper_limit)
        if int(self[ATTR_SENSOR_TYPE][VALUE]) in self.linear_sensor_type:
            out = self.command(u":SPL{},{},{}\n".format(self.channel, int(value), int(upper_limit)), with_token=True)
        else:
            revolution = 360 * 10 ** abs(self[ATTR_ENCODER_RESOLUTION][VALUE])
            min_revolutions = int(value // revolution)
            min_angular_position = int(value - min_revolutions * revolution)
            max_revolutions = int(upper_limit // revolution)
            max_angular_position = int(upper_limit - max_revolutions * revolution)
            out = self.command(
                u":SAL{},{},{},{},{}\n".format(self.channel, int(min_angular_position), int(min_revolutions),
                                               int(max_angular_position), int(max_revolutions)), callback,
                with_token=True)

        self.set_raw_value([ATTR_POSITION, MIN], value)
        self.get_position_limit_range()
        return out

    def set_upper_position_limit(self, value, callback=None):
        self.logger.info(u"Set upper position limit to {}".format(value))
        lower_limit = self[ATTR_POSITION_LOWER_LIMIT][VALUE]
        # self.set_value(ATTR_POSITION_UPPER_LIMIT, int(value))
        if lower_limit is None:
            lower_limit = value
            # self.set_value(ATTR_POSITION_LOWER_LIMIT, lower_limit)
        if int(self[ATTR_SENSOR_TYPE][VALUE]) in self.linear_sensor_type:
            out = self.command(u":SPL{},{},{}\n".format(self.channel, int(value), int(lower_limit)), callback,
                               with_token=True)
        else:
            revolution = 360 * 10 ** abs(self[ATTR_ENCODER_RESOLUTION][VALUE])
            max_revolutions = int(value // revolution)
            max_angular_position = int(value - max_revolutions * revolution)
            min_revolutions = int(lower_limit // revolution)
            min_angular_position = int(lower_limit - min_revolutions * revolution)
            out = self.command(
                u":SAL{},{},{},{},{}\n".format(self.channel, int(min_angular_position), int(min_revolutions),
                                               int(max_angular_position), int(max_revolutions)), callback,
                with_token=True)

        self.set_raw_value([ATTR_POSITION, MAX], value)
        self.get_position_limit_range()
        return out

    def set_step_frequency(self, value, callback=None):
        self.logger.info(u"Set step frequency to {} Hz".format(value))
        return self.command(u":SCLF{},{}\n".format(self.channel, int(value)), callback, with_token=True)

    def set_step_amplitude(self, value):
        self.logger.info(u"Set step amplitude to {}".format(value))
        self[ATTR_STEP_AMPLITUDE][VALUE] = value

    def set_velocity(self, value, callback=None):
        self.logger.info(u"Set velocity to {}".format(value))
        return self.command(u":SCLS{},{}\n".format(self.channel, int(value)), callback, with_token=True)

    def set_acceleration(self, value, callback=None):
        self.logger.info(u"Set acceleration to {}".format(value))
        return self.command(u":SCLA{},{}\n".format(self.channel, int(value)), callback, with_token=True)

    def set_safe_direction(self, value, callback=None):
        self.logger.info(u"Set safe direction to {}".format(value))
        value = self.safe_direction_modes[value]
        return self.command(u":SSD{},{}\n".format(self.channel, value), callback, with_token=True)
